import paraview
from paraview import servermanager
from paraview.util import proxy as proxy_util

from paraview.simple.session import GetActiveSource
from paraview.simple import GetAnimationScene


# ==============================================================================
# Extracts and Extractors
# ==============================================================================
def CreateExtractor(name, proxy=None, registrationName=None):
    """Creates a new extractor and returns it.

    :param name: The type of the extractor to create.
    :type name: `str`
    :param proxy: The proxy to generate the extract from. If not specified
        the active source is used.
    :type proxy: :class:`paraview.servermanager.Proxy`, optional
    :param registrationName: The registration name to use to register
        the extractor with the ProxyManager. If not specified a unique name
        will be generated.
    :type name: `str`, optional

    :return: The extractor created, on success, else None
    :rtype: :class:`paraview.servermanager.Proxy` or `None`

    """
    if proxy is None:
        proxy = GetActiveSource()
    if not proxy:
        raise RuntimeError("Could determine input for extractor")

    if proxy.Port > 0:
        rawProxy = proxy.SMProxy.GetOutputPort(proxy.Port)
    else:
        rawProxy = proxy.SMProxy

    controller = servermanager.vtkSMExtractsController()

    # Special case of 'steering' extractors : add it to the steerable extractors list
    # managed by the in situ helper class, not to the extracts controller, because
    # they are not associated with a writer.
    from paraview.catalyst.detail import IsInsitu, RegisterExtractor

    if name == "steering" and IsInsitu():
        rawExtractor = controller.CreateSteeringExtractor(rawProxy, registrationName)
        extractor = servermanager._getPyProxy(rawExtractor)
        UpdateSteerableParameters(extractor, registrationName)
    else:
        rawExtractor = controller.CreateExtractor(rawProxy, name, registrationName)
        extractor = servermanager._getPyProxy(rawExtractor)
        if IsInsitu():
            # tag the extractor to know which pipeline this came from.
            RegisterExtractor(extractor)
    return extractor


def GetExtractors(proxy=None):
    """Returns a list of extractors associated with the proxy.

    :param proxy: The proxy to return the extractors associated with.
                  If not specified (or is None) then all extractors are returned.
    :type proxy: :class:`paraview.servermanager.Proxy`, optional

    :return: List of associated extractors if any. May return an empty list.
    :rtype: list of :class:`paraview.servermanager.Proxy`

    """
    controller = servermanager.vtkSMExtractsController()
    pxm = servermanager.ProxyManager()
    extractors = pxm.GetProxiesInGroup("extractors").values()
    if proxy is None:
        return list(extractors)
    else:
        return [
            g for g in extractors if controller.IsExtractor(g.SMProxy, proxy.SMProxy)
        ]


def FindExtractor(registrationName):
    """
    Returns an extractor with a specific registrationName.

    :param registrationName: Registration name for the extractor.
    :type registrationName: `str`

    :return: The extractor or None
    :rtype: :class:`paraview.servermanager.Proxy` or `None`
    """
    pxm = servermanager.ProxyManager()
    return pxm.GetProxy("extractors", registrationName)


def SaveExtractsUsingCatalystOptions(options):
    """Generate extracts using a :class:`CatalystOptions` object. This is intended for
    use when a Catalyst analysis script is being run in batch mode.

    :param options: Catalyst options proxy
    :type options: :class:`CatalystOptions`
    """
    # handle the case where options is CatalystOptions
    return SaveExtracts(
        ExtractsOutputDirectory=options.ExtractsOutputDirectory,
        GenerateCinemaSpecification=options.GenerateCinemaSpecification,
    )


def SaveExtracts(**kwargs):
    """Generate extracts. Parameters are forwarded for 'SaveAnimationExtracts'
    Currently, supported keyword parameters are:

    :param ExtractsOutputDirectory: root directory for extracts. If the directory
        does not exist, it will be created.
    :type ExtractsOutputDirectory: `str`
    :param GenerateCinemaSpecification: If set to `True`, enables generation of a
        cinema specification.
    :type GenerateCinemaSpecification: bool
    :param FrameRate: The frame rate in frames-per-second if the extractor
        supports it.
    :type FrameRate: int
    :param FrameStride: The number of timesteps to skip after a timestep is saved.
    :type FrameStride: int
    :param FrameWindow: The range of timesteps to extract.
    :type FrameWindow: 2-element tuple or list of ints
    """
    # currently, Python state files don't have anything about animation,
    # however, we rely on animation to generate the extracts, so we ensure
    # that the animation is updated based on timesteps in data by explicitly
    # calling UpdateAnimationUsingDataTimeSteps.
    scene = GetAnimationScene()
    if not scene:
        from paraview import print_warning

        print_warning("This Edition does not support animations. Cannot save extracts.")
        return False

    scene.UpdateAnimationUsingDataTimeSteps()

    controller = servermanager.ParaViewPipelineController()
    proxy = servermanager.misc.SaveAnimationExtracts()
    controller.PreInitializeProxy(proxy)
    proxy.AnimationScene = scene
    controller.PostInitializeProxy(proxy)
    proxy_util.set(proxy, **kwargs)
    return proxy.SaveExtracts()


def CreateSteerableParameters(
    steerable_proxy_type_name,
    steerable_proxy_registration_name="SteeringParameters",
    result_mesh_name="steerable",
):
    """Creates a steerable proxy for Catalyst use cases.

    :param steerable_proxy_type_name: Name of the proxy type to create.
    :type steerable_proxy_type_name: str
    :param steerable_proxy_registration_name: Registration name of the proxy
        created by this function. If not provided, defaults to "SteeringParameters".
    :type steerable_proxy_registration_name: str
    :param result_mesh_name: The name of the resulting mesh. If not provided,
        defaults to "steerable".
    :type result_mesh_name: str
    :return: Proxy of the specified type.
    :rtype: Steerable proxy."""
    pxm = servermanager.ProxyManager()
    steerable_proxy = pxm.NewProxy("sources", steerable_proxy_type_name)
    pxm.RegisterProxy("sources", steerable_proxy_registration_name, steerable_proxy)
    steerable_proxy_wrapper = servermanager._getPyProxy(steerable_proxy)
    UpdateSteerableParameters(steerable_proxy_wrapper, result_mesh_name)
    return steerable_proxy_wrapper


def UpdateSteerableParameters(steerable_proxy, steerable_source_name):
    """Updates a steerable proxy of a provided source name.

    :param steerable_proxy: The steerable proxy to update.
    :type steerable_proxy: Source proxy.
    :param steerable_source_name: The name of the steerable source proxy.
    :type steerable_source_name: str"""
    helper = paraview.modules.vtkPVInSitu.vtkInSituInitializationHelper
    helper.UpdateSteerableParameters(steerable_proxy.SMProxy, steerable_source_name)



from . import layer
from . import channel
import json
import os

#
# imageview class
#
class imageview:
    """ImageView Class

    A collection of settings that define a specific way of compositing the
    elements of an image. Once it is set up, the imageview's layers can
    be iterated over to return an order-dependent set of data plus colormaps
    that can be composited.
    """

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, color):
        if (len(color) == 3) and (all(isinstance(value, (float, int)) for value in color)):
            self._background = color
        else:
            print("ERROR: color arg must be three values, each either an int or float") 
            self._background = [0.0, 0.0, 0.0]

    @property
    def use_depth(self):
        return self._use_depth

    @use_depth.setter
    def use_depth(self, value):
        self._use_depth = value

    @property
    def use_shadow(self):
        return self._use_shadow

    @use_shadow.setter
    def use_shadow(self, value):
        self._use_shadow = value

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value):
        self._depth = value

    @property
    def shadow(self):
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        self._shadow = value

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value):
        self._image = value

    @property
    def dims(self):
        return self._dims

    @dims.setter
    def dims(self, value):
        self._dims = value

    def __init__(self, cview):
        self.active_layers = []
        self.active_channels = {} 
            # a CIS view of a cinema database
        self.cisview = cview
        self.data = {}
        self.use_depth = False
        self.use_shadow = False
        self.background = [0.0, 0.0, 0.0]

    def get_active_layers(self):
        return self.active_layers

    def activate_layer(self, layer):
        if not layer in self.active_layers:
            self.active_layers.append(layer)

    def deactivate_layer(self, layer):
        if layer in self.active_layers:
            self.active_layers.remove(layer)

    def is_active_layer(self, layer):
        return layer in self.active_layers

    #
    # will activate a channel in an inactive layer
    #
    def activate_channel(self, layer, channel):
        self.active_channels[layer] = channel

    def get_active_channel_data(self, layer):
        channel = self.get_active_channel(layer) 
        data = self.cisview.get_image(self.image).get_layer(layer).get_channel(channel).data
        return data 

    def get_layer(self, layer):
        # TODO: error check
        return self.data[self.image]

    def get_channel(self, layername):
        # TODO: error check
        return self.data[layername].channel

    def get_layer_dims(self, layer):
        return self.cisview.get_image(self.image).get_layer(layer).dims 

    def get_layer_offset(self, layer):
        return self.cisview.get_image(self.image).get_layer(layer).offset 

    def get_active_channel(self, layer):
        results = None

        if layer in self.active_channels:
            results = self.active_channels[layer] 

        return results    

    def get_variable_range(self, layer):
        var = self.cisview.get_variable(self.get_active_channel(layer))

        data = [var["min"], var["max"]] 
        if var["type"] == "float":
            data = [float(var["min"]), float(var["max"])]
        elif var["type"] == "float":
            data = [int(var["min"]), int(var["max"])]

        return data 

    def update(self):
        imdata = self.cisview.get_image_parameters()
        self.dims   = imdata["dims"]

        # TODO: error if image not set
        for l in self.active_layers:
            ldata = self.cisview.get_layer_parameters(self.image, l)
            newlayer = layer.layer(l)
            newlayer.name = l
            newlayer.dims = ldata["dims"]
            newlayer.offset = ldata["offset"]
            self.data[l] = newlayer

            extract = self.cisview.get_channel_extract(self.image, l, self.active_channels[l])
            cdata = self.cisview.get_channel_parameters(self.image, l, self.active_channels[l])
            newchannel = channel.channel()
            newchannel.name = self.active_channels[l]
            newchannel.load(extract[0])
            newchannel.colormap = self.load_colormap(cdata["colormap"])
            newchannel.var      = cdata["variable"]["name"]
            newchannel.vartype  = cdata["variable"]["type"]
            newchannel.varmin   = cdata["variable"]["min"]
            newchannel.varmax   = cdata["variable"]["max"]
            newlayer.channel = newchannel

            # load the depth map
            if self.use_depth:
                extract = self.cisview.get_channel_extract(self.image, l, "CISDepth") 
                newchannel = channel.channel()
                newchannel.name = "CISDepth" 
                newchannel.load(extract[0])
                newlayer.depth = newchannel

            # load the shadow map
            if self.use_shadow:
                extract = self.cisview.get_channel_extract(self.image, l, "CISShadow") 
                # did the data load?
                # this is equivalent to asking if the channel is there
                # TODO: find a better way to express requesting a load, but getting
                #       no data, i.e. the channel is not there
                if extract:
                    newchannel = channel.channel()
                    newchannel.name = "CISShadow"
                    newchannel.load(extract[0])
                    newlayer.shadow = newchannel

    def get_layer_data(self):
        return self.data

    def load_colormap(self, params):
        # return a default gray colormap if nothing else
        colormap = {
                    "colorspace" : "rgb",
                    "name" : "default",
                    "points" : [{'x': 0.0, 'r': 0.0, 'g': 0.0, 'b': 0.0, 'a': 1.0},
                                {'x': 1.0, 'r': 1.0, 'g': 1.0, 'b': 1.0, 'a': 1.0},
                               ]
                   }

        if "type" in params:
            # for now, parse as a local file
            # TODO: add logic to detect and load remote URLs
            if params["type"] is "url":
                # this path is currently required to be a ParaView json colormap
                # local to the cinema database 
                fullpath = os.path.join(self.cisview.cdb.path, params["url"])

                colormap["colorspace"] = "rgb"
                colormap["points"] = []
                with open(fullpath) as cmfile:
                    data = json.load(cmfile)
                    numpoints = int(len(data[0]["RGBPoints"])/4)
                    for i in range(numpoints):
                        colormap["points"].append(
                                                   {'x': data[0]["RGBPoints"][i*4],
                                                    'r': data[0]["RGBPoints"][i*4+1],
                                                    'g': data[0]["RGBPoints"][i*4+2],
                                                    'b': data[0]["RGBPoints"][i*4+3],
                                                    'a': 1.0 
                                                   }
                                                 )

        return colormap





from paraview import servermanager


def set(proxy, **props):
    """
    Sets one or more properties of the given proxy.
    Pass in arguments of the form `property_name=value` to this function to
    set property values.

    For example::

        set(
            proxy, # CAN NOT be None
            Center=[1, 2, 3],
            Radius=3.5,
        )

    :param proxy: The pipeline source whose properties should be set.
    :type proxy: Source proxy
    :param params: A variadic list of `key=value` pairs giving values of
        specific named properties in the pipeline source. For a list of available
        properties, call `help(proxy)`.
    """
    if proxy is None:
        raise RuntimeError("Proxy can not be None")

    proxy = servermanager._getPyProxy(proxy)
    for k, v in props.items():
        proxy.__setattr__(k, v)


def rename(proxy, group=None, name=None):
    """
    Renames the given proxy. This is the name used by :func:`FindSource` and
    is displayed in the Pipeline Browser.

    :param proxy: The proxy to be renamed
    :type proxy: Proxy object
    :param group: The group in which the proxy lives. Can be retrieved with
        `proxy.GetXMLGroup()` but not always (layouts vs misc).
    :type group: str
    :param newName: The new name of the proxy.
    :type newName: str
    """
    pxm = servermanager.ProxyManager()

    if group is None:
        group = proxy.GetXMLGroup()

    old_name = pxm.GetProxyName(group, proxy)
    if old_name and name != old_name:
        pxm.RegisterProxy(group, name, proxy)
        pxm.UnRegisterProxy(group, old_name, proxy)


def unregister(proxy):
    """
    Unregister proxy from ParaView Pipeline Controller
    """
    controller = servermanager.ParaViewPipelineController()
    controller.UnRegisterProxy(proxy)



from paraview import servermanager

from paraview.simple.session import GetActiveSource, GetActiveView
from paraview.simple.proxy import GetSources

# ==============================================================================
# Animation methods
# ==============================================================================


def GetTimeKeeper():
    """Returns the timekeeper proxy for the active session. Timekeeper is often used
    to manage time step information known to the ParaView application.

    :return: timekeeper
    :rtype: :class:`paraview.servermanager.TimeKeeper`"""
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    controller = servermanager.ParaViewPipelineController()
    return controller.FindTimeKeeper(session)


def GetAnimationScene():
    """Returns the application-wide animation scene. ParaView has only one global
    animation scene. This method provides access to that. You are free to create
    additional animation scenes directly, but those scenes won't be shown in the
    ParaView GUI.

    :return: The animation scene
    :rtype: :class:`paraview.servermanager.AnimationScene`"""
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    controller = servermanager.ParaViewPipelineController()
    return controller.GetAnimationScene(session)


# -----------------------------------------------------------------------------


def AnimateReader(reader=None, view=None):
    """This is a utility function that, given a reader and a view
    animates over all time steps of the reader.

    :param reader: The reader source to iterate over. Optional, defaults
        to the active source.
    :type reader: reader proxy
    :param view: The view to animate. Optional, defaults to the active view.
    :type view: view proxy
    :return: A new animation scene used to play the animation.
    :rtype: :class:`paraview.simple.AnimationSceneProxy`"""
    if not reader:
        reader = GetActiveSource()
    if not view:
        view = GetActiveView()

    return servermanager.AnimateReader(reader, view)


# -----------------------------------------------------------------------------


def GetRepresentationAnimationHelper(sourceproxy):
    """Method that returns the representation animation helper for a
    source proxy. It creates a new one if none exists."""
    # ascertain that proxy is a source proxy
    if not sourceproxy in GetSources().values():
        return None
    for proxy in servermanager.ProxyManager():
        if proxy.GetXMLName() == "RepresentationAnimationHelper" and proxy.GetProperty(
            "Source"
        ).IsProxyAdded(sourceproxy.SMProxy):
            return proxy
    # helper must have been created during RegisterPipelineProxy().
    return None


# -----------------------------------------------------------------------------


def GetAnimationTrack(propertyname_or_property, index=None, proxy=None, registrationName=None):
    """Returns an animation cue for the property. If one doesn't exist then a
    new one will be created.
    Typical usage::

      track = GetAnimationTrack("Center", 0, sphere) or
      track = GetAnimationTrack(sphere.GetProperty("Radius")) or

      # this returns the track to animate visibility of the active source in
      # all views.
      track = GetAnimationTrack("Visibility")

    For animating properties on implicit planes etc., use the following
    signatures::

      track = GetAnimationTrack(slice.SliceType.GetProperty("Origin"), 0) or
      track = GetAnimationTrack("Origin", 0, slice.SliceType)
    """
    if not proxy:
        proxy = GetActiveSource()
    if not isinstance(proxy, servermanager.Proxy):
        raise TypeError("proxy must be a servermanager.Proxy instance")
    if isinstance(propertyname_or_property, str):
        propertyname = propertyname_or_property
    elif isinstance(propertyname_or_property, servermanager.Property):
        prop = propertyname_or_property
        propertyname = prop.Name
        proxy = prop.Proxy
    else:
        raise TypeError(
            "propertyname_or_property must be a string or servermanager.Property"
        )

    # To handle the case where the property is actually a "display" property, in
    # which case we are actually animating the "RepresentationAnimationHelper"
    # associated with the source.
    if (
        propertyname in ["Visibility", "Opacity"]
        and proxy.GetXMLName() != "RepresentationAnimationHelper"
    ):
        proxy = GetRepresentationAnimationHelper(proxy)
    if not proxy or not proxy.GetProperty(propertyname):
        raise AttributeError("Failed to locate property %s" % propertyname)

    scene = GetAnimationScene()
    for cue in scene.Cues:
        try:
            if cue.AnimatedProxy == proxy and cue.AnimatedPropertyName == propertyname:
                if index == None or index == cue.AnimatedElement:
                    return cue
        except AttributeError:
            pass

    # matching animation track wasn't found, create a new one.
    cue = servermanager.animation.KeyFrameAnimationCue()
    cue.AnimatedProxy = proxy
    cue.AnimatedPropertyName = propertyname
    if index != None:
        cue.AnimatedElement = index
    session = servermanager.ActiveConnection.Session
    pxm = servermanager.ProxyManager(session)
    pxm.RegisterProxy("animation", registrationName, cue)
    scene.Cues.append(cue)
    return cue


# -----------------------------------------------------------------------------


def GetCameraTrack(view=None, registrationName=None):
    """Returns the camera animation track for the given view.

    :param view: The view whose camera animation track should be retrieved.
                 If no view is specified, the active view will be used. If no existing camera
                 animation track is found, a new one will be created.
    :param registrationName: Optional name to register the camera animation cue with.
    :type view: View proxy
    :return: Camera animation cue
    :rtype: :class:`paraview.servermanager.CameraAnimationCue`

    """
    if not view:
        view = GetActiveView()
    if not view:
        raise ValueError("No view specified")
    scene = GetAnimationScene()
    for cue in scene.Cues:
        if cue.AnimatedProxy == view and cue.GetXMLName() == "CameraAnimationCue":
            return cue
    # no cue was found, create a new one.
    cue = servermanager.animation.CameraAnimationCue()
    cue.AnimatedProxy = view
    session = servermanager.ActiveConnection.Session
    pxm = servermanager.ProxyManager(session)
    pxm.RegisterProxy("animation", registrationName, cue)
    scene.Cues.append(cue)
    return cue


# -----------------------------------------------------------------------------


def GetTimeTrack():
    """Returns the animation track used to control the time requested from all
    sources during playback

    :return: The time animation track.
    :rtype: :class:`paraview.servermanager.TimeAnimationCue`"""
    scene = GetAnimationScene()
    if not scene:
        raise RuntimeError("Missing animation scene")
    controller = servermanager.ParaViewPipelineController()
    return controller.GetTimeAnimationTrack(scene)



from paraview import servermanager

from paraview.simple.session import active_objects, GetActiveSource
from paraview.simple.rendering import GetRepresentation


# ==============================================================================
# Lookup Table / Scalarbar methods
# ==============================================================================
# -----------------------------------------------------------------------------
def HideUnusedScalarBars(view=None):
    """Hides all unused scalar bars from the view. A scalar bar is used if some
    data is shown in that view that is coloring using the transfer function
    shown by the scalar bar.

    :param view: View in which unused scalar bars should be hidden. Optional,
        defaults to the active view.
    :type view: View proxy"""
    if not view:
        view = active_objects.view
    if not view:
        raise ValueError("'view' argument cannot be None with no active is present.")
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    return tfmgr.UpdateScalarBars(view.SMProxy, tfmgr.HIDE_UNUSED_SCALAR_BARS)


def HideScalarBarIfNotNeeded(lut, view=None):
    """Hides the given scalar bar if it is not used by any of the displayed data.

    :param lut: The lookup table (lut) or scalar bar proxy
    :type lut: Scalar bar proxy
    :param view: The view in which the scalar bar should be hidden. Optional,
        defaults to hiding scalar bars in the active view."""
    if not view:
        view = active_objects.view
    if not view:
        raise ValueError("'view' argument cannot be None with no active present.")
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    return tfmgr.HideScalarBarIfNotNeeded(lut.SMProxy, view.SMProxy)


def UpdateScalarBars(view=None):
    """Hides all unused scalar bars and shows used scalar bars. A scalar bar is used
    if some data is shown in that view that is coloring using the transfer function
    shown by the scalar bar.

    :param view: The view in which scalar bar visibility should be changed. Optional,
        defaults to using the active view.
    :type view: View proxy"""
    if not view:
        view = active_objects.view
    if not view:
        raise ValueError("'view' argument cannot be None with no active is present.")
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    return tfmgr.UpdateScalarBars(
        view.SMProxy, tfmgr.HIDE_UNUSED_SCALAR_BARS | tfmgr.SHOW_USED_SCALAR_BARS
    )


def UpdateScalarBarsComponentTitle(ctf, representation=None):
    """Update the component portion of the title in all scalar bars using the provided
    lookup table. The representation is used to recover the array from which the
    component title was obtained.

    :param ctf: The lookup table that the scalar bar represents. Optional, defaults to
        the representation of the active source in the active view.
    :type ctf: Transfer function proxy.
    :param representation: If provided, it is the representation to use to recover
        the array that provides the component title portion. Optional, defaults to the
        active representation.
    :type representation: Representation proxy
    :return: `True` if operation succeeded, `False` otherwise.
    :rtype: bool"""
    if not representation:
        view = active_objects.view
        proxy = active_objects.source
        if not view:
            raise ValueError(
                "'representation' argument cannot be None with no active view."
            )
        if not proxy:
            raise ValueError(
                "'representation' argument cannot be None with no active source."
            )
        representation = GetRepresentation(view, proxy)
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    return tfmgr.UpdateScalarBarsComponentTitle(ctf.SMProxy, representation.SMProxy)


def GetScalarBar(ctf, view=None):
    """Returns the scalar bar for the given color transfer function in the given view.

    :param ctf: The color transfer function proxy whose scalar bar representation should
        be returned.
    :type ctf: Color transfer function proxy.
    :param view: View from which the scalar bar proxy should be retrieved.
        Optional, defaults to the active view, if possible.
    :type view: View proxy.
    :return: The scalar bar proxy for the color transfer function if found. his will
        either return an existing scalar bar or create a new one.
    :rtype: Scalar bar proxy"""
    view = view if view else active_objects.view
    if not view:
        raise ValueError(
            "'view' argument cannot be None when no active view is present"
        )
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    sb = servermanager._getPyProxy(
        tfmgr.GetScalarBarRepresentation(ctf.SMProxy, view.SMProxy)
    )
    return sb


# ==============================================================================
# Miscellaneous functions.
# ==============================================================================
def ShowInteractiveWidgets(proxy=None):
    """If possible in the current environment, this function will request the
    application to show the interactive widget(s) for the given proxy.

    :param proxy: The proxy whose associated interactive widgets should be shown.
        Optional, if not provided the active source's widgets are shown.
    :type proxy: Source proxy."""
    proxy = proxy if proxy else GetActiveSource()
    if not proxy:
        raise ValueError("No 'proxy' was provided and no active source was found.")
    _InvokeWidgetUserEvent(proxy, "ShowWidget")


def HideInteractiveWidgets(proxy=None):
    """If possible in the current environment, this function will
    request the application to hide the interactive widget(s) for the given proxy.

    :param proxy: The proxy whose associated interactive widgets should be hidden.
        Optional, if not provided the active source's widgets are hidden.
    :type proxy: Source proxy."""
    proxy = proxy if proxy else GetActiveSource()
    if not proxy:
        raise ValueError("No 'proxy' was provided and no active source was found.")
    _InvokeWidgetUserEvent(proxy, "HideWidget")


def _InvokeWidgetUserEvent(proxy, event):
    """Internal method used by ShowInteractiveWidgets/HideInteractiveWidgets"""
    if proxy:
        proxy.InvokeEvent("UserEvent", event)
        # Since in 5.0 and earlier, ShowInteractiveWidgets/HideInteractiveWidgets
        # was called with the proxy being the filter proxy (eg. Clip) and not the
        # proxy that has the widget i.e. (Clip.ClipType), we explicitly handle it
        # by iterating of proxy list properties and then invoking the event on
        # their value proxies too.
        for smproperty in proxy:
            if smproperty.FindDomain("vtkSMProxyListDomain"):
                _InvokeWidgetUserEvent(smproperty.GetData(), event)



# -*- coding: utf-8 -*-
"""Activate venv for current interpreter:

Use `from paraview.web import venv` along one of the following
 - `--venv /path/to/venv/base` argument
 - environment variable `PV_VENV=/path/to/venv/base`

This can be used when you must use an existing Python interpreter, not the venv bin/python.
"""
import os
import site
import sys

VENV_BASE = None
VENV_LOADED = False

if "--venv" in sys.argv:
    VENV_BASE = os.path.abspath(sys.argv[sys.argv.index("--venv") + 1])

if os.environ.get("PV_VENV"):
    VENV_BASE = os.path.abspath(os.environ.get("PV_VENV"))

if not VENV_LOADED and VENV_BASE and os.path.exists(VENV_BASE):
    VENV_LOADED = True
    # Code inspired by virutal-env::bin/active_this.py
    bin_dir = os.path.join(VENV_BASE, "bin")
    os.environ["PATH"] = os.pathsep.join([bin_dir] + os.environ.get("PATH", "").split(os.pathsep))
    os.environ["VIRTUAL_ENV"] = VENV_BASE
    prev_length = len(sys.path)

    if sys.platform == "win32":
        python_libs = os.path.join(VENV_BASE, "Lib/site-packages")
    else:
        python_libs = os.path.join(VENV_BASE, f"lib/python{sys.version_info.major}.{sys.version_info.minor}/site-packages")

    site.addsitedir(python_libs)
    sys.path[:] = sys.path[prev_length:] + sys.path[0:prev_length]
    sys.real_prefix = sys.prefix
    sys.prefix = VENV_BASE
    #
    print(f"ParaView is using venv: {VENV_BASE}")



r"""private module with utilities to be used by modules in this package alone"""


def compare(test_image, baseline_image):
    from vtkmodules.vtkTestingRendering import vtkTesting
    testing = vtkTesting()
    testing.AddArgument("-V")
    testing.AddArgument(baseline_image)
    if testing.RegressionTest(test_image, 0.05) == vtkTesting.FAILED:
        raise RuntimeError("Regression test failed!")
    return True



"""
Atmosphere Data - Readers and Filters for ParaView
===================================================

This module provides NetCDF readers and filters for E3SM/EAM (Energy Exascale
Earth System Model / E3SM Atmosphere Model) data visualization in ParaView.

Dependencies
------------
- netCDF4: For reading NetCDF climate data files
- numpy: For numerical array operations
- math: Standard library for spherical coordinate transformations

No external dependencies like pyproj or pandas are required.

Readers
-------
EAMDataReader
    Multi-output reader that produces three separate outputs:
    - Port 0: Surface variables (e.g., surface temperature, precipitation)
    - Port 1: Middle layer variables on lev levels as hexahedral volume mesh
    - Port 2: Interface layer variables on ilev levels as hexahedral volume mesh

Filters
-------
EAMProjectToSphere
    Projects lat/lon data onto a 3D sphere for globe visualization.
    Supports scaling for vertical exaggeration of atmospheric layers.

Data Format
-----------
EAM/E3SM data consists of:

- Data files (.nc): NetCDF files containing time-varying atmospheric variables
- Connectivity files (.nc): NetCDF files defining the unstructured mesh topology
  with corner lat/lon coordinates for each grid cell

The readers automatically detect:

- Horizontal dimensions by matching sizes between connectivity and data files
- Vertical dimensions (lev for mid-levels, ilev for interface levels)
- Time dimensions for animation support
- Fill values (_FillValue or missing_value attributes)

Vertical Coordinates
--------------------
EAM uses hybrid sigma-pressure coordinates. The readers can compute pressure
levels from hybrid coefficients (hyam, hybm, hyai, hybi) if explicit level
coordinates are not provided.

Usage Example
-------------
In ParaView:

1. Use File -> Open to select a NetCDF data file
2. Choose "EAM Data Reader"
3. Set the connectivity file path in the Properties panel
4. Select variables to load from the array selection widgets
5. Apply to visualize

For spherical projection:

1. Apply EAMDataReader reader
2. Apply EAMProjectToSphere filter
3. Adjust scale factor for vertical exaggeration if needed
"""

import math

from vtkmodules.numpy_interface import dataset_adapter as dsa
from vtkmodules.vtkCommonCore import vtkPoints, vtkDataArraySelection
from vtkmodules.vtkCommonDataModel import vtkUnstructuredGrid, vtkCellArray
from vtkmodules.vtkFiltersCore import vtkAppendFilter
from vtkmodules.util import vtkConstants, numpy_support
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from paraview import print_error, print_warning

try:
    import netCDF4
    import numpy as np
    _has_deps = True
except ImportError as ie:
    print_error(
        "Missing required Python modules/packages. Algorithms in this module may "
        "not work as expected! \n {0}".format(ie))
    _has_deps = False


# ==============================================================================
# Constants and Helper Classes
# ==============================================================================

class AtmosphereConstants:
    """Constants for EAM atmospheric model hybrid coordinates."""
    LEV = 'lev'
    HYAM = 'hyam'
    HYBM = 'hybm'
    ILEV = 'ilev'
    HYAI = 'hyai'
    HYBI = 'hybi'
    P0 = float(1e5)
    PS0 = float(1e5)


class DimMeta:
    """Stores dimension metadata from NetCDF files."""

    def __init__(self, name, size, data=None):
        self.name = name
        self.size = size
        self.long_name = None
        self.units = None
        self.data = data

    def update_from_variable(self, var_info):
        """Update metadata from netCDF variable info."""
        try:
            self.long_name = var_info.getncattr('long_name')
        except AttributeError:
            pass
        try:
            self.units = var_info.getncattr('units')
        except AttributeError:
            pass

    def __repr__(self):
        return f"DimMeta(name='{self.name}', size={self.size})"


class VarMeta:
    """Stores variable metadata from NetCDF files."""

    def __init__(self, name, info, horizontal_dim=None):
        self.name = name
        self.dimensions = info.dimensions
        self.fillval = np.nan
        self.long_name = None
        self.horizontal_dim = horizontal_dim

        # Determine if transpose needed (horizontal dim not last)
        if horizontal_dim and horizontal_dim in self.dimensions:
            self.transpose = self.dimensions.index(horizontal_dim) != len(self.dimensions) - 1
        else:
            self.transpose = False

        # Extract fill value - try both _FillValue and missing_value
        for attr in ['_FillValue', 'missing_value']:
            try:
                self.fillval = info.getncattr(attr)
                break
            except (AttributeError, KeyError):
                pass

        # Get long_name if available
        try:
            self.long_name = info.getncattr('long_name')
        except (AttributeError, KeyError):
            pass

    def has_dim(self, dim_name):
        """Check if variable has a specific dimension."""
        return dim_name in self.dimensions

    def get_vertical_dim(self):
        """Return the vertical dimension (lev or ilev) if present."""
        if AtmosphereConstants.LEV in self.dimensions:
            return AtmosphereConstants.LEV
        elif AtmosphereConstants.ILEV in self.dimensions:
            return AtmosphereConstants.ILEV
        return None

    def __repr__(self):
        return f"VarMeta(name='{self.name}', dims={self.dimensions})"


# ==============================================================================
# Helper Functions
# ==============================================================================

def compare(data, arrays, dim):
    """Compare arrays for hybrid coordinate validation."""
    ref = data[arrays[0]][:].flatten()
    if len(ref) != dim:
        raise Exception("Length of hya_/hyb_ variable does not match the corresponding dimension")
    for array in arrays[1:]:
        comp = data[array][:].flatten()
        if not np.array_equal(ref, comp):
            return None
    return ref


def FindSpecialVariable(data, lev, hya, hyb):
    """Find or compute level coordinates from hybrid coefficients."""
    dim = data.dimensions.get(lev, None)
    if dim is None:
        raise Exception(f'{lev} not found in dimensions')
    dim = dim.size
    var = np.array(list(data.variables.keys()))

    if lev in var:
        return data[lev][:].flatten()

    _hyai = var[np.char.find(var, hya) != -1]
    _hybi = var[np.char.find(var, hyb) != -1]
    if len(_hyai) != len(_hybi):
        raise Exception('Unmatched pair of hya and hyb variables found')

    p0 = AtmosphereConstants.P0
    ps0 = AtmosphereConstants.PS0

    if len(_hyai) == 1:
        hyai = data[_hyai[0]][:].flatten()
        hybi = data[_hybi[0]][:].flatten()
        if not (len(hyai) == dim and len(hybi) == dim):
            raise Exception('Lengths of arrays for hya_ and hyb_ variables do not match')
        return ((hyai * p0) + (hybi * ps0)) / 100.0
    else:
        hyai = compare(data, _hyai, dim)
        hybi = compare(data, _hybi, dim)
        if hyai is None or hybi is None:
            raise Exception('Values within hya_ and hyb_ arrays do not match')
        return ((hyai * p0) + (hybi * ps0)) / 100.0


def createModifiedCallback(anobject):
    """Create a weak-reference callback for ParaView modified events."""
    import weakref
    weakref_obj = weakref.ref(anobject)
    anobject = None

    def _markmodified(*args, **kwargs):
        o = weakref_obj()
        if o is not None:
            o.Modified()
    return _markmodified


def identify_horizontal_dim(meshdata, vardata):
    """
    Identify horizontal dimension from connectivity file and match in data file.
    Returns (conn_dim_name, data_dim_name, size) or (None, None, None) on failure.
    """
    conn_dims = list(meshdata.dimensions.keys())
    if not conn_dims:
        return None, None, None

    # First dimension in connectivity file is typically the horizontal dim
    conn_dim = conn_dims[0]
    conn_size = meshdata.dimensions[conn_dim].size

    # Find matching dimension in data file by size
    for dim_name, dim_obj in vardata.dimensions.items():
        if dim_obj.size == conn_size:
            return conn_dim, dim_name, conn_size

    return conn_dim, None, conn_size


def find_lat_lon_vars(meshdata):
    """Find lat/lon variable names in connectivity file."""
    mvars = np.array(list(meshdata.variables.keys()))
    lat_matches = mvars[np.char.find(mvars, 'corner_lat') > -1]
    lon_matches = mvars[np.char.find(mvars, 'corner_lon') > -1]
    if len(lat_matches) == 0 or len(lon_matches) == 0:
        return None, None
    return lat_matches[0], lon_matches[0]


def build_2d_geometry(lat, lon, ncells):
    """Build VTK geometry arrays for 2D mesh."""
    coords = np.empty((len(lat), 3), dtype=np.float64)
    coords[:, 0] = lon
    coords[:, 1] = lat
    coords[:, 2] = 0.0

    vtk_coords = vtkPoints()
    vtk_coords.SetData(dsa.numpyTovtkDataArray(coords))

    cell_types = np.empty(ncells, dtype=np.uint8)
    cell_types.fill(vtkConstants.VTK_QUAD)
    cell_types = numpy_support.numpy_to_vtk(
        num_array=cell_types, deep=True, array_type=vtkConstants.VTK_UNSIGNED_CHAR)

    offsets = np.arange(0, (4 * ncells) + 1, 4, dtype=np.int64)
    offsets = numpy_support.numpy_to_vtk(
        num_array=offsets, deep=True, array_type=vtkConstants.VTK_ID_TYPE)

    cells = np.arange(ncells * 4, dtype=np.int64)
    cells = numpy_support.numpy_to_vtk(
        num_array=cells, deep=True, array_type=vtkConstants.VTK_ID_TYPE)

    cell_array = vtkCellArray()
    cell_array.SetData(offsets, cells)

    return vtk_coords, cell_types, cell_array


def build_3d_volume_geometry(lat, lon, levels, ncells2d):
    """
    Build VTK geometry arrays for 3D volumetric mesh with hexahedral cells.

    Creates hexahedra by connecting adjacent vertical layers. Each hex cell
    spans from level N to level N+1, resulting in (num_levels - 1) layers
    of hexahedral cells.

    Parameters
    ----------
    lat : ndarray
        Flattened latitude coordinates for cell corners.
    lon : ndarray
        Flattened longitude coordinates for cell corners.
    levels : ndarray
        Vertical level coordinates.
    ncells2d : int
        Number of 2D cells per horizontal layer.

    Returns
    -------
    vtk_coords : vtkPoints
        Points for all vertical levels.
    cell_types : vtkUnsignedCharArray
        Cell type array (all VTK_HEXAHEDRON).
    cell_array : vtkCellArray
        Cell connectivity array.
    """
    nlev = len(levels)
    npts_per_level = len(lat)

    # Build 3D coordinates for all levels
    coords = np.empty((nlev, npts_per_level, 3), dtype=np.float64)
    for i, z in enumerate(levels):
        coords[i, :, 0] = lon
        coords[i, :, 1] = lat
        coords[i, :, 2] = z
    coords = coords.reshape(nlev * npts_per_level, 3)

    vtk_coords = vtkPoints()
    vtk_coords.SetData(dsa.numpyTovtkDataArray(coords))

    # Build hexahedral cells connecting adjacent levels
    # Number of hex layers = nlev - 1
    hex_layers = nlev - 1
    numhexes = ncells2d * hex_layers

    # Build hex connectivity efficiently using numpy
    # Each hex connects 4 points from lower level to 4 points from upper level
    hexconn = np.empty((hex_layers, ncells2d * 8), dtype=np.int64)
    for i in range(hex_layers):
        # Lower level point indices (4 corners per cell)
        lower = np.arange(i * npts_per_level, (i + 1) * npts_per_level, dtype=np.int64)
        lower = lower.reshape(ncells2d, 4).transpose()
        # Upper level point indices (4 corners per cell)
        upper = np.arange((i + 1) * npts_per_level, (i + 2) * npts_per_level, dtype=np.int64)
        upper = upper.reshape(ncells2d, 4).transpose()
        # Combine: 4 lower + 4 upper = 8 points per hex
        conn = np.append(lower, upper, axis=0).flatten('F')
        hexconn[i] = conn
    hexconn = hexconn.flatten()

    cell_types = np.empty(numhexes, dtype=np.uint8)
    cell_types.fill(vtkConstants.VTK_HEXAHEDRON)
    cell_types = numpy_support.numpy_to_vtk(
        num_array=cell_types, deep=True, array_type=vtkConstants.VTK_UNSIGNED_CHAR)

    offsets = np.arange(0, (8 * numhexes) + 1, 8, dtype=np.int64)
    offsets = numpy_support.numpy_to_vtk(
        num_array=offsets, deep=True, array_type=vtkConstants.VTK_ID_TYPE)

    cells = numpy_support.numpy_to_vtk(
        num_array=hexconn, deep=True, array_type=vtkConstants.VTK_ID_TYPE)

    cell_array = vtkCellArray()
    cell_array.SetData(offsets, cells)

    return vtk_coords, cell_types, cell_array


def load_variable_data(vardata, varmeta, time_idx):
    """Load variable data for a given time index, handling transpose and fill values."""
    try:
        if varmeta.transpose:
            data = vardata[varmeta.name][:].data[time_idx].transpose().flatten()
        else:
            data = vardata[varmeta.name][:].data[time_idx].flatten()
        return np.where(data == varmeta.fillval, np.nan, data)
    except Exception as e:
        print_error(f"Error loading variable {varmeta.name}: {e}")
        return np.array([])


def load_variable_data_averaged(vardata, varmeta, time_idx, ncells2d, nlev):
    """
    Load 3D variable data and average between adjacent vertical levels.

    For hexahedral cells that span between levels N and N+1, the cell data
    is computed as the average of the values at those two levels.

    Parameters
    ----------
    vardata : netCDF4.Dataset
        NetCDF dataset containing the variable.
    varmeta : VarMeta
        Variable metadata.
    time_idx : int
        Time index to load.
    ncells2d : int
        Number of 2D cells per horizontal layer.
    nlev : int
        Number of vertical levels in the data.

    Returns
    -------
    ndarray
        Averaged cell data with shape (nlev-1) * ncells2d.
    """
    try:
        if varmeta.transpose:
            data = vardata[varmeta.name][:].data[time_idx].transpose().flatten()
        else:
            data = vardata[varmeta.name][:].data[time_idx].flatten()

        # Replace fill values with NaN
        data = np.where(data == varmeta.fillval, np.nan, data)

        # Average between adjacent levels
        hex_layers = nlev - 1
        averaged = np.empty(hex_layers * ncells2d, dtype=np.float64)

        for i in range(hex_layers):
            lower = data[i * ncells2d:(i + 1) * ncells2d]
            upper = data[(i + 1) * ncells2d:(i + 2) * ncells2d]
            averaged[i * ncells2d:(i + 1) * ncells2d] = (lower + upper) / 2.0

        return averaged
    except Exception as e:
        print_error(f"Error loading variable {varmeta.name}: {e}")
        return np.array([])


def process_point_to_sphere(point, max_z, radius, scale):
    """Convert lat/lon/z point to spherical coordinates."""
    theta = point[0]
    phi = 90 - point[1]
    rho = (max_z - point[2]) * scale + radius if point[2] != 0 else radius
    x = rho * math.sin(math.radians(phi)) * math.cos(math.radians(theta))
    y = rho * math.sin(math.radians(phi)) * math.sin(math.radians(theta))
    z = rho * math.cos(math.radians(phi))
    return [x, y, z]


# ==============================================================================
# EAMDataReader - Multi-output reader for 2D and 3D volumetric data
# ==============================================================================

class EAMDataReader(VTKPythonAlgorithmBase):
    """
    Multi-output NetCDF reader for E3SM/EAM atmospheric model data.

    This reader produces three separate outputs for different variable types:
    - Port 0 (Surface): Surface variables with 2 dimensions (time, ncol) as quad mesh
    - Port 1 (Middle Layer): Variables on middle levels as hexahedral volume mesh
    - Port 2 (Interface Layer): Variables on interface levels as hexahedral volume mesh

    The 3D outputs are converted to volumetric data directly in the reader,
    producing hexahedral cells that span between adjacent vertical levels.
    This enables immediate volume rendering without requiring additional filters.

    The reader automatically detects horizontal dimensions by matching the connectivity
    file's cell count with dimensions in the data file. It handles fill values
    (_FillValue, missing_value) by converting them to NaN.

    Parameters
    ----------
    DataFile : str
        Path to the NetCDF data file containing atmospheric variables.
    ConnectivityFile : str
        Path to the NetCDF connectivity file containing mesh topology
        (corner_lat, corner_lon arrays defining cell vertices).

    Outputs
    -------
    Port 0 : vtkUnstructuredGrid
        Surface mesh with quad cells. Each cell corresponds to one
        horizontal grid column. Contains selected surface variables as cell data.
    Port 1 : vtkUnstructuredGrid
        Volumetric mesh for middle layer (lev) variables with hexahedral cells.
        Cell data is averaged between adjacent levels. Contains 'lev' and 'numlev'
        in field data.
    Port 2 : vtkUnstructuredGrid
        Volumetric mesh for interface layer (ilev) variables with hexahedral cells.
        Cell data is averaged between adjacent levels. Contains 'ilev' and 'numilev'
        in field data.

    Notes
    -----
    - Vertical coordinates are computed from hybrid sigma-pressure coefficients
      (hyam, hybm for lev; hyai, hybi for ilev) if explicit level arrays are missing.
    - The reader caches NetCDF file handles to avoid repeated file opens.
    - Variables are grouped by type in the ParaView UI for easy selection.
    - Time dimension is automatically detected and exposed for animation.
    - 3D outputs have (num_levels - 1) hexahedral cells in the vertical direction
      since each hex spans two adjacent levels.

    Example
    -------
    In ParaView Python shell::

        from paraview.simple import EAMDataReader
        reader = EAMDataReader()
        reader.DataFile = '/path/to/data.nc'
        reader.ConnectivityFile = '/path/to/connectivity.nc'
        reader.SurfaceVariables = ['TREFHT']
        reader.UpdatePipeline()
    """

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self, nInputPorts=0, nOutputPorts=3, outputType='vtkUnstructuredGrid')

        self._data_filename = None
        self._conn_filename = None
        self._time_steps = []

        self._horizontal_dim = None
        self._data_horizontal_dim = None
        self._dimensions = {}
        self._variables = {}

        self._surface_selection = vtkDataArraySelection()
        self._middle_layer_selection = vtkDataArraySelection()
        self._interface_layer_selection = vtkDataArraySelection()

        for sel in [self._surface_selection, self._middle_layer_selection, self._interface_layer_selection]:
            sel.AddObserver("ModifiedEvent", createModifiedCallback(self))

        self._mesh_dataset = None
        self._var_dataset = None
        self._cached_mesh_filename = None
        self._cached_var_filename = None
        self._area_var = None

    def __del__(self):
        self._close_datasets()

    def _close_datasets(self):
        for ds in [self._mesh_dataset, self._var_dataset]:
            if ds is not None:
                try:
                    ds.close()
                except Exception:
                    pass
        self._mesh_dataset = None
        self._var_dataset = None

    def _get_mesh_dataset(self):
        if self._conn_filename != self._cached_mesh_filename or self._mesh_dataset is None:
            if self._mesh_dataset is not None:
                try:
                    self._mesh_dataset.close()
                except Exception:
                    pass
            self._mesh_dataset = netCDF4.Dataset(self._conn_filename, "r")
            self._cached_mesh_filename = self._conn_filename
        return self._mesh_dataset

    def _get_var_dataset(self):
        if self._data_filename != self._cached_var_filename or self._var_dataset is None:
            if self._var_dataset is not None:
                try:
                    self._var_dataset.close()
                except Exception:
                    pass
            self._var_dataset = netCDF4.Dataset(self._data_filename, "r")
            self._cached_var_filename = self._data_filename
        return self._var_dataset

    def _clear(self):
        self._variables.clear()
        self._dimensions.clear()
        self._time_steps.clear()
        self._horizontal_dim = None
        self._data_horizontal_dim = None
        self._area_var = None

    def _populate_metadata(self):
        if self._data_filename is None or self._conn_filename is None:
            return

        meshdata = self._get_mesh_dataset()
        vardata = self._get_var_dataset()

        self._horizontal_dim, self._data_horizontal_dim, _ = identify_horizontal_dim(meshdata, vardata)
        if not self._data_horizontal_dim:
            print_warning("Could not match horizontal dimension in data file")
            return

        self._surface_selection.RemoveAllArrays()
        self._middle_layer_selection.RemoveAllArrays()
        self._interface_layer_selection.RemoveAllArrays()

        for dim_name, dim_obj in vardata.dimensions.items():
            dim_meta = DimMeta(dim_name, dim_obj.size)
            if dim_name in vardata.variables:
                dim_meta.update_from_variable(vardata.variables[dim_name])
                try:
                    dim_meta.data = vardata[dim_name][:].data
                except Exception:
                    pass
            self._dimensions[dim_name] = dim_meta

        for name, info in vardata.variables.items():
            if self._data_horizontal_dim not in info.dimensions:
                continue

            varmeta = VarMeta(name, info, self._data_horizontal_dim)
            self._variables[name] = varmeta

            if len(info.dimensions) == 1 and 'area' in name.lower():
                self._area_var = varmeta
                continue

            ndims = len(info.dimensions)
            if ndims == 2:
                self._surface_selection.AddArray(name)
            elif ndims == 3:
                if AtmosphereConstants.LEV in info.dimensions:
                    self._middle_layer_selection.AddArray(name)
                elif AtmosphereConstants.ILEV in info.dimensions:
                    self._interface_layer_selection.AddArray(name)

        self._surface_selection.DisableAllArrays()
        self._middle_layer_selection.DisableAllArrays()
        self._interface_layer_selection.DisableAllArrays()

        if 'time' in vardata.variables:
            self._time_steps = list(vardata['time'][:].data.flatten())

    def SetDataFileName(self, fname):
        if fname and fname != "None" and fname != self._data_filename:
            self._data_filename = fname
            self._clear()
            self._populate_metadata()
            self.Modified()

    def SetConnectivityFileName(self, fname):
        if fname != self._conn_filename:
            self._conn_filename = fname
            self._clear()
            if self._data_filename:
                self._populate_metadata()
            self.Modified()

    def GetTimestepValues(self):
        return self._time_steps

    def GetSurfaceDataArrays(self):
        return self._surface_selection

    def GetMiddleLayerDataArrays(self):
        return self._middle_layer_selection

    def GetInterfaceLayerDataArrays(self):
        return self._interface_layer_selection

    def RequestInformation(self, request, inInfo, outInfo):
        executive = self.GetExecutive()
        for i in range(3):
            port = outInfo.GetInformationObject(i)
            port.Remove(executive.TIME_STEPS())
            port.Remove(executive.TIME_RANGE())
            if self._time_steps:
                for t in self._time_steps:
                    port.Append(executive.TIME_STEPS(), t)
                port.Append(executive.TIME_RANGE(), self._time_steps[0])
                port.Append(executive.TIME_RANGE(), self._time_steps[-1])
        return 1

    def _get_time_index(self, outInfo, port_index):
        executive = self.GetExecutive()
        time_info = outInfo.GetInformationObject(port_index)
        if time_info.Has(executive.UPDATE_TIME_STEP()) and len(self._time_steps) > 1:
            time = time_info.Get(executive.UPDATE_TIME_STEP())
            for i, t in enumerate(self._time_steps):
                if time <= t:
                    return i
        return 0

    def RequestData(self, request, inInfo, outInfo):
        if not self._conn_filename or not self._data_filename:
            print_error("Data file or connectivity file not provided!")
            return 0

        if not _has_deps:
            print_error("Required Python module 'netCDF4' or 'numpy' missing!")
            return 0

        meshdata = self._get_mesh_dataset()
        vardata = self._get_var_dataset()

        if not self._data_horizontal_dim:
            print_error("Could not identify horizontal dimension")
            return 0

        lat_var, lon_var = find_lat_lon_vars(meshdata)
        if not lat_var or not lon_var:
            print_error("Could not find lat/lon variables in connectivity file")
            return 0

        lat = meshdata[lat_var][:].data.flatten()
        lon = meshdata[lon_var][:].data.flatten()
        ncells2D = meshdata.dimensions[self._horizontal_dim].size

        executive = self.GetExecutive()
        from_port = request.Get(executive.FROM_OUTPUT_PORT())
        time_idx = self._get_time_index(outInfo, from_port)

        # Output 0: Surface data
        outputSurface = dsa.WrapDataObject(vtkUnstructuredGrid.GetData(outInfo, 0))
        vtk_coords, cell_types, cell_array = build_2d_geometry(lat, lon, ncells2D)
        outputSurface.SetPoints(vtk_coords)
        outputSurface.VTKObject.SetCells(cell_types, cell_array)

        for name, varmeta in self._variables.items():
            if self._surface_selection.ArrayIsEnabled(name):
                data = load_variable_data(vardata, varmeta, time_idx)
                outputSurface.CellData.append(data, name)

        # Output 1: Middle layer (lev) - volumetric hexahedral mesh
        try:
            lev = FindSpecialVariable(vardata, AtmosphereConstants.LEV, AtmosphereConstants.HYAM, AtmosphereConstants.HYBM)
            if lev is not None:
                nlev = len(lev)
                outputMiddle = dsa.WrapDataObject(vtkUnstructuredGrid.GetData(outInfo, 1))
                vtk_coords, cell_types, cell_array = build_3d_volume_geometry(lat, lon, lev, ncells2D)
                outputMiddle.SetPoints(vtk_coords)
                outputMiddle.VTKObject.SetCells(cell_types, cell_array)

                for name, varmeta in self._variables.items():
                    if self._middle_layer_selection.ArrayIsEnabled(name):
                        data = load_variable_data_averaged(vardata, varmeta, time_idx, ncells2D, nlev)
                        outputMiddle.CellData.append(data, name)

                outputMiddle.FieldData.append(nlev - 1, "numlev")
                outputMiddle.FieldData.append(lev, "lev")
        except Exception as e:
            print_error(f"Error processing middle layer variables: {e}")

        # Output 2: Interface layer (ilev) - volumetric hexahedral mesh
        try:
            ilev = FindSpecialVariable(vardata, AtmosphereConstants.ILEV, AtmosphereConstants.HYAI, AtmosphereConstants.HYBI)
            if ilev is not None:
                nilev = len(ilev)
                outputInterface = dsa.WrapDataObject(vtkUnstructuredGrid.GetData(outInfo, 2))
                vtk_coords, cell_types, cell_array = build_3d_volume_geometry(lat, lon, ilev, ncells2D)
                outputInterface.SetPoints(vtk_coords)
                outputInterface.VTKObject.SetCells(cell_types, cell_array)

                for name, varmeta in self._variables.items():
                    if self._interface_layer_selection.ArrayIsEnabled(name):
                        data = load_variable_data_averaged(vardata, varmeta, time_idx, ncells2D, nilev)
                        outputInterface.CellData.append(data, name)

                outputInterface.FieldData.append(nilev - 1, "numilev")
                outputInterface.FieldData.append(ilev, "ilev")
        except Exception as e:
            print_error(f"Error processing interface layer variables: {e}")

        return 1


# ==============================================================================
# EAMProjectToSphere - Project data onto spherical coordinates
# ==============================================================================

class EAMProjectToSphere(VTKPythonAlgorithmBase):
    """
    Project lat/lon data onto a 3D sphere for globe visualization.

    This filter transforms points from geographic coordinates (longitude as X,
    latitude as Y, vertical level as Z) into spherical coordinates for
    rendering data on a 3D globe. It supports vertical exaggeration through
    a scale factor to make atmospheric layers visible.

    The transformation maps:

    - Longitude -> azimuthal angle (theta)
    - Latitude -> polar angle (phi), measured from north pole
    - Z-coordinate -> radial distance from sphere center

    Parameters
    ----------
    Data Layer : bool
        If True, adds a small offset (1 unit) to the radius. This is useful
        when overlaying data on a separate sphere geometry to prevent
        z-fighting artifacts. Default is False.
    Scale : float
        Vertical exaggeration factor for the Z-coordinate. Higher values make
        vertical structure more visible. Default is 1.0. A value of 0 places
        all points on the sphere surface.

    Input
    -----
    vtkUnstructuredGrid or vtkPolyData
        Mesh with points in geographic coordinates.

        - X: Longitude in degrees (-180 to 180 or 0 to 360)
        - Y: Latitude in degrees (-90 to 90)
        - Z: Vertical level (e.g., pressure in hPa, or index)

    Output
    ------
    vtkUnstructuredGrid
        Same topology as input but with points transformed to spherical
        coordinates. Point data and cell data are preserved unchanged.

    Notes
    -----
    - The default sphere radius is 2000 units, designed for visualization
      with typical atmospheric data scales.
    - For 2D surface data (Z=0), all points are placed exactly on the sphere.
    - For 3D data, higher Z values (e.g., higher pressure = lower altitude)
      result in smaller radii, placing those points closer to the sphere center.
    - vtkPolyData input is automatically converted to vtkUnstructuredGrid.

    Example
    -------
    In ParaView Python shell::

        from paraview.simple import *
        reader = EAMDataReader()
        reader.DataFile = '/path/to/data.nc'
        reader.ConnectivityFile = '/path/to/connectivity.nc'
        reader.SurfaceVariables = ['TREFHT']
        reader.Update()

        # Project surface output onto sphere
        sphere = EAMProjectToSphere(Input=OutputPort(reader, 0))
        sphere.DataLayer = True  # Slight offset for layering
        sphere.Update()

        # For middle layer data with vertical exaggeration
        sphere3d = EAMProjectToSphere(Input=OutputPort(reader, 1))
        sphere3d.Scale = 10.0  # Exaggerate vertical structure
        sphere3d.Update()
    """

    def __init__(self):
        super().__init__(nInputPorts=1, nOutputPorts=1, outputType="vtkUnstructuredGrid")
        self._is_data_layer = False
        self._radius = 2000
        self._scale = 1.0

    def FillInputPortInformation(self, port, info):
        info.Set(self.INPUT_REQUIRED_DATA_TYPE(), "vtkDataSet")
        return 1

    def SetDataLayer(self, is_data):
        if self._is_data_layer != is_data:
            self._is_data_layer = is_data
            self.Modified()

    def SetScalingFactor(self, scale):
        if self._scale != float(scale):
            self._scale = float(scale)
            self.Modified()

    def RequestDataObject(self, request, inInfo, outInfo):
        inData = self.GetInputData(inInfo, 0, 0)
        outData = self.GetOutputData(outInfo, 0)
        assert inData is not None
        if outData is None or (not outData.IsA(inData.GetClassName())):
            outData = inData.NewInstance()
            outInfo.GetInformationObject(0).Set(outData.DATA_OBJECT(), outData)
        return super().RequestDataObject(request, inInfo, outInfo)

    def RequestData(self, request, inInfo, outInfo):
        if not _has_deps:
            print_error("Required Python module 'numpy' missing!")
            return 0

        inData = self.GetInputData(inInfo, 0, 0)
        outData = self.GetOutputData(outInfo, 0)

        # Handle PolyData input by converting to UnstructuredGrid
        if inData.IsA('vtkPolyData'):
            afilter = vtkAppendFilter()
            afilter.AddInputData(inData)
            afilter.Update()
            outData.DeepCopy(afilter.GetOutput())
        else:
            outData.DeepCopy(inData)

        inWrap = dsa.WrapDataObject(inData)
        outWrap = dsa.WrapDataObject(outData)

        inPoints = np.array(inWrap.Points)
        max_z = np.max(inPoints[:, 2])

        # Offset radius for data layer
        radius = (self._radius + 1) if self._is_data_layer else self._radius

        # Transform points to spherical coordinates
        outPoints = np.array([
            process_point_to_sphere(pt, max_z, radius, self._scale)
            for pt in inPoints
        ])

        vtk_coords = vtkPoints()
        vtk_coords.SetData(numpy_support.numpy_to_vtk(
            outPoints, deep=True, array_type=vtkConstants.VTK_FLOAT))
        outWrap.SetPoints(vtk_coords)

        return 1



"""
Execute example by running the following commad:

   $ pvpython -m paraview.simple.example.demo1

"""

from paraview.simple import *


def demo1():
    """
    Simple demo that create the following pipeline::

       sphere - shrink +
       cone            + > append

    """
    # Create a sphere of radius = 2, theta res. = 32
    # This object becomes the active source.
    ss = Sphere(Radius=2, ThetaResolution=32)

    # Apply the shrink filter. The Input property is optional. If Input
    # is not specified, the filter is applied to the active source.
    shr = Shrink(Input=ss)

    # Create a cone source.
    cs = Cone()

    # Append cone and shrink
    app = AppendDatasets()
    app.Input = [shr, cs]

    # Show the output of the append filter. The argument is optional
    # as the app filter is now the active object.
    Show(app)

    # Render the default view.
    Render()

    # Lock the script execution to interact with 3D view
    Interact()


if __name__ == "__main__":
    demo1()



from paraview.simple import *
from vtkmodules.vtkCommonMath import vtkMatrix4x4


def SetModelTransformMatrix(matrix):
    rvProxy = GetActiveView()
    invmat = vtkMatrix4x4()
    invmat.DeepCopy(matrix)
    invmat.Invert()

    mElts = [matrix.GetElement(i, j) for i in range(4) for j in range(4)]
    rvProxy.SetPropertyWithName("ModelTransformMatrix", mElts)

    mElts = [invmat.GetElement(i, j) for i in range(4) for j in range(4)]
    rvProxy.SetPropertyWithName("PhysicalToWorldMatrix", mElts)

    rvProxy.UpdateVTKObjects()


class PythonInteractorBase():
    def Initialize(self, vtkSelf):
        pass

    def Update(self, vtkSelf):
        pass

    def HandleTracker(self, vtkSelf, role, sensor, matrix):
        pass

    def HandleButton(self, vtkSelf, role, button, state):
        pass

    def HandleValuator(self, vtkSelf, numChannels, channelData):
        pass



from paraview import servermanager

# ==============================================================================
# Materials.
# ==============================================================================


def GetMaterialLibrary():
    """Returns the material library for the active session.

    :return: The material library
    :rtype: :class:`paraview.servermanager.MaterialLibrary`"""
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    controller = servermanager.ParaViewPipelineController()
    return controller.FindMaterialLibrary(session)


# ==============================================================================
# Textures.
# ==============================================================================
def CreateTexture(filename=None, trivial_producer_key=None, proxyname=None):
    """Creates and returns a new ImageTexture proxy.
    The texture is not attached to anything by default but it can be applied
    to things, for example the view, like so:

        GetActiveView().UseTexturedBackground = 1
        GetActiveView().BackgroundTexture = CreateTexture("foo.png")

    :param filename: The name of the image file to load as a texture. It can be
        `None`, in which case the texture is read from a trivial producer identified
        by the `trivial_producer_key`
    :type filename: str
    :param trivial_producer_key: Identifier of the texture image source on the
        server. This is mainly used by scene importers when `filename` is `None`.
    :type trivial_producer_key: str
    :param proxyname: The name for the texture when it is registered in ParaView.
    :type proxyname: str
    :return: The newly created texture.
    :rtype: Image texture proxy.
    """
    pxm = servermanager.ProxyManager()
    textureproxy = pxm.NewProxy("textures", "ImageTexture")
    controller = servermanager.ParaViewPipelineController()
    if filename is not None:
        controller.SMController.RegisterTextureProxyFromFile(textureproxy, filename, proxyname)
    else:
        controller.SMController.RegisterTextureProxy(
            textureproxy, trivial_producer_key, proxyname)

    return servermanager._getPyProxy(textureproxy)


def FindTextureOrCreate(registrationName, filename=None, trivial_producer_key=None):
    """Finds or creates a new ImageTexture proxy.

    :param registrationName: The name for the texture when it is registered in ParaView.
    :type registrationName: str
    :param filename: Path to the texture image source file. Optional, if not provided,
        the `trivial_producer_key` should be provided to tell which image source to use
        for the texture.
    :type filename: str
    :param trivial_producer_key: Identifier of the texture image source on the
        server. This is mainly used by scene importers when `filename` is `None`.
    :type trivial_producer_key: str"""
    pxm = servermanager.ProxyManager()
    textureproxy = pxm.GetProxy("textures", registrationName)
    if textureproxy is None:
        return CreateTexture(filename, trivial_producer_key, registrationName)
    else:
        return textureproxy



# SPDX-FileCopyrightText: Copyright (c) Kitware Inc.
# SPDX-License-Identifier: BSD-3-Clause

"""selection is a module for defining and using selections via Python. It provides
a convenience layer to provide functionality provided by the underlying C++ classes.

A simple example::

.. code:: python

    from paraview.selection import *

    # Create a new sphere proxy on the active connection and register it
    # in the sources group.
    sphere = Sphere(ThetaResolution=16, PhiResolution=32)

    # Show the sphere
    Show(sphere)

    # Select cells falling inside a rectangle defined by two points on a diagonal,
    # (300, 400) and (600, 800).
    SelectSurfaceCells(Rectangle=[300, 400, 600, 800])

    # Add cells on processor 0 with ids 1 and 2 to the current selection.
    SelectIDs(IDs=[0, 1, 0, 2], FieldType='CELL', Source=sphere, Modifier='ADD')

    # Remove cells on processor 0 with ids 4 and 5 from the current selection.
    SelectIDS(IDs=[0, 4, 0, 5], FieldType='CELL', Source=sphere, Modifier='SUBTRACT')

    # Extract the currently extracted cells
    es = ExtractSelection()
    Show(es)

"""

from __future__ import absolute_import, division, print_function

import paraview
from paraview import servermanager as sm
import paraview.simple


class SelectionProxy(sm.Proxy):
    """ Special proxy wrapper type for Selections.
    """

    def __init__(self, **args):
        super(SelectionProxy, self).__init__(**args)


def _createSelection(proxyname, **args):
    """ Utility function to create a selection source. Basically a factory function
    that creates a proxy of a given name and hands off keyword arguments to the newly
    created proxy.
    """
    session = sm.ActiveConnection.Session
    pxm = sm.ProxyManager(session)

    groupname = 'filters' if proxyname == 'AppendSelections' else 'sources'
    proxy = pxm.NewProxy(groupname, proxyname)

    s = SelectionProxy(proxy=proxy)
    for name, value in args.items():
        # Warning: this will add only the attributes initially passed in when creating the proxy.
        s.add_attribute(name, value)
        s.SetPropertyWithName(name, value)

    return s


def CreateSelection(proxyname, registrationname, groupname='selection_sources', **args):
    """Make a new selection source proxy. Can be either vtkSelection or vtkAppendSelection.
    Use this so that selection don't show up in the pipeline."""

    s = _createSelection(proxyname, **args)

    session = sm.ActiveConnection.Session
    pxm = sm.ProxyManager(session)

    pxm.RegisterProxy(groupname, registrationname, s)
    return sm._getPyProxy(s)


def _collectSelectionPorts(selectedReps, selectionSources, SelectBlocks=False, Modifier=None):
    """ Utility to collect all the selection output ports from a list of selected representations
    and sources. Selects blocks instead of cells if the SelectBlocks option is True.
    The Modifier can be used to modify the currently existing selection through one of the values
    'ADD', 'SUBTRACT', 'TOGGLE' or None if the new selection should replace the existing one.

    """
    outputPorts = []

    if not selectedReps or selectedReps.GetNumberOfItems() <= 0:
        return outputPorts

    if not selectionSources or selectionSources.GetNumberOfItems() <= 0:
        return outputPorts

    if selectedReps.GetNumberOfItems() != selectionSources.GetNumberOfItems():
        return outputPorts

    for i in range(0, selectedReps.GetNumberOfItems()):
        repr = selectedReps.GetItemAsObject(i)
        selectionSource = selectionSources.GetItemAsObject(i)

        if not repr:
            continue

        # Ensure selected representation is registered with the proxy manager
        pxm = repr.GetSessionProxyManager()
        if not pxm:
            return

        # Get the output port from the representation input
        inputProperty = repr.GetProperty("Input")
        selectedDataSource = inputProperty.GetProxy(0)
        portNumber = inputProperty.GetOutputPortForConnection(0)

        outputPort = selectedDataSource.GetOutputPort(portNumber)

        prevAppendSelections = selectedDataSource.GetSelectionInput(portNumber)

        # Convert block selection from index-based selection
        from paraview.modules.vtkRemotingViews import vtkSMSelectionHelper
        import paraview.vtk as vtk
        if SelectBlocks:
            dinfo = selectedDataSource.GetDataInformation()
            targetContentType = vtk.vtkSelectionNode.BLOCK_SELECTORS if dinfo.DataSetTypeIsA(
                "vtkPartitionedDataSetCollection") else vtk.vtkSelectionNode.BLOCKS
            selectionSource = vtkSMSelectionHelper.ConvertSelectionSource(targetContentType, selectionSource,
                                                                          selectedDataSource, portNumber)

        # create an append-selections proxy with the selection source as the only input
        newAppendSelections = vtkSMSelectionHelper.NewAppendSelectionsFromSelectionSource(selectionSource)

        # Handle selection modifier
        if Modifier == 'ADD':
            vtkSMSelectionHelper.AddSelection(prevAppendSelections, newAppendSelections)
        elif Modifier == 'SUBTRACT':
            vtkSMSelectionHelper.SubtractSelection(prevAppendSelections, newAppendSelections)
        elif Modifier == 'TOGGLE':
            vtkSMSelectionHelper.ToggleSelection(prevAppendSelections, newAppendSelections)
        else:
            vtkSMSelectionHelper.IgnoreSelection(prevAppendSelections, newAppendSelections)

        selectedDataSource.SetSelectionInput(portNumber, newAppendSelections, outputPort.GetPortIndex())

        if SelectBlocks:
            # A new selection proxy was allocated when converting to a block
            # selection, so we need to delete unregister it for garbage collection
            selectionSource.UnRegister(None)
        newAppendSelections.UnRegister(None)

        # Add output port to list of ports and return it
        outputPorts.append(outputPort)

    return outputPorts


def _surfaceSelectionHelper(rectOrPolygon, view, type, Modifier=None):
    """ Selects mesh elements on a surface

    - rectOrPolygon - represents either a rectangle or polygon. If a rectangle, consists of a list
      containing the bottom left (x1, y1) and top right (x2, y2) corner of a rectangle defining the
      selection region in the format [x1, y1, x2, y2]. If a polygon, list of points defining the
      polygon in x-y pairs (e.g., [x1, y1, x2, y2, x3, y3,...])
    - view - the view in which to make the selection
    - type - the type of mesh element: 'POINT', 'CELL', or 'BLOCK'
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
      should modify the existing selection.
    """
    if not view.SMProxy.IsA('vtkSMRenderViewProxy'):
        return

    # ensure the view has rendered so that the selection is valid
    paraview.simple.Render(view)

    if len(rectOrPolygon) == 0:
        rectOrPolygon = [0, 0, view.ViewSize[0], view.ViewSize[1]]

    from paraview.vtk import vtkCollection, vtkIntArray

    selectedReps = vtkCollection()
    selectionSources = vtkCollection()

    if type.startswith('POLYGON'):
        polygon = vtkIntArray()
        polygon.SetNumberOfComponents(2)

        for component in rectOrPolygon:
            polygon.InsertNextValue(component)

    if type == 'POINT':
        view.SelectSurfacePoints(rectOrPolygon, selectedReps, selectionSources, 0)
    elif type == 'CELL' or type == 'BLOCK':
        view.SelectSurfaceCells(rectOrPolygon, selectedReps, selectionSources, 0)
    elif type == 'FRUSTUM_POINTS':
        view.SelectFrustumPoints(rectOrPolygon, selectedReps, selectionSources, 0)
    elif type == 'FRUSTUM_CELLS':
        view.SelectFrustumCells(rectOrPolygon, selectedReps, selectionSources, 0)
    elif type == 'POLYGON_POINTS':
        view.SelectPolygonPoints(polygon, selectedReps, selectionSources, 0)
    elif type == 'POLYGON_CELLS':
        view.SelectPolygonCells(polygon, selectedReps, selectionSources, 0)
    else:
        raise RuntimeError("Invalid type %s" % type)

    _collectSelectionPorts(selectedReps, selectionSources, type == 'BLOCK', Modifier)

    paraview.simple.Render(view)


def SelectSurfacePoints(Rectangle=[], Polygon=[], View=None, Modifier=None):
    """Select visible points within a rectangular or polygon region.

    - Rectangle - list containing bottom left (x1, y1) and top right (x2, y2) corner of a
      rectangle defining the selection region in the format [x1, y1, x2, y2]. Defined in pixels.
      If not empty, the Polygon parameter must be empty or None.
    - Polygon - list of 2D points defining a polygon in which visible points should be selected,
       e.g., [x1, y1, x2, y2, ..., xn, yn]. Defined in pixels. If not empty, the Rectangle parameter
       must be empty or None.
    - View - the view in which to perform the selection. If None, uses the current active view.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
      should modify the existing selection.
    """
    if not View:
        View = paraview.simple.GetActiveView()

    if Rectangle and Polygon:
        raise RuntimeError("Can only have one of Rectangle or Polygon named parameters set.")
    elif Rectangle:
        return _surfaceSelectionHelper(Rectangle, View, 'POINT', Modifier)
    elif Polygon:
        return _surfaceSelectionHelper(Polygon, View, 'POLYGON_POINTS', Modifier)
    else:
        raise RuntimeError("Need to set either the Rectangle or Polygon named parameters.")

    return None


def SelectSurfaceCells(Rectangle=[], Polygon=[], View=None, Modifier=None):
    """Select visible cells within a rectangular or polygon region.

    - Rectangle - list containing bottom left (x1, y1) and top right (x2, y2) corner of a
      rectangle defining the selection region in the format [x1, y1, x2, y2]. Defined in pixels.
      If not empty, the Polygon parameter must be empty or None.
    - Polygon - list of 2D points defining a polygon in which visible points should be selected,
       e.g., [x1, y1, x2, y2, ..., xn, yn]. Defined in pixels. If not empty, the Rectangle parameter
       must be empty or None.
    - View - the view in which to perform the selection. If None, uses the current active view.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
      should modify the existing selection.
    """
    if not View:
        View = paraview.simple.GetActiveView()

    if Rectangle and Polygon:
        raise RuntimeError("Can only have one of Rectangle or Polygon named parameters set.")
    if Rectangle:
        return _surfaceSelectionHelper(Rectangle, View, 'CELL', Modifier)
    elif Polygon:
        return _surfaceSelectionHelper(Polygon, View, 'POLYGON_CELLS', Modifier)
    else:
        raise RuntimeError("")


def SelectSurfaceBlocks(Rectangle=[], View=None, Modifier=None):
    """Select visible blocks within a rectangular region.

    - Rectangle - list containing bottom left (x1, y1) and top right (x2, y2) corner of a
      rectangle defining the selection region in the format [x1, y1, x2, y2]. Defined in pixels.
    - View - the view in which to perform the selection. If None, uses the current active view.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
      should modify the existing selection.
    """
    if not View:
        View = paraview.simple.GetActiveView()

    if not Modifier:
        # Clear the current selection - work around default behavior or utility function, which is
        # to add blocks.
        ClearSelection()

        # need to trigger render since otherwise the representation for the
        # selection that was just cleared may not have updated.
        paraview.simple.Render(View)

    return _surfaceSelectionHelper(Rectangle, View, 'BLOCK', Modifier)


def SelectPointsThrough(Rectangle=[], View=None, Modifier=None):
    """Select all points within a rectangular region regardless of their visibility.

    - Rectangle - list containing bottom left (x1, y1) and top right (x2, y2) corner of a
      rectangle defining the selection region in the format [x1, y1, x2, y2]. Defined in pixels.
    - View - the view in which to perform the selection. If None, uses the current active view.
    """
    if not View:
        View = paraview.simple.GetActiveView()

    return _surfaceSelectionHelper(Rectangle, View, 'FRUSTUM_POINTS', Modifier)


def SelectCellsThrough(Rectangle=[], View=None, Modifier=None):
    """Select all cells within a rectangular region regardless of their visibility.

    - Rectangle - list containing bottom left (x1, y1) and top right (x2, y2) corner of a
      rectangle defining the selection region in the format [x1, y1, x2, y2]. Defined in pixels.
    - View - the view in which to perform the selection. If None, uses the current active view.
    """
    if not View:
        View = paraview.simple.GetActiveView()

    return _surfaceSelectionHelper(Rectangle, View, 'FRUSTUM_CELLS', Modifier)


def _selectIDsHelper(proxyname, IDs=[], FieldType='POINT', ContainingCells=False, Source=None, Modifier=None):
    """Selects IDs of a given field type.

    - proxyname - name of the selection proxy to instantiate
    - IDs - list of IDs of attribute types to select.
    - FieldType - type of attribute to select, e.g., 'POINT', 'CELL'
    - ContainingCells - if True and FieldType is 'POINT', select the cells containing the
        points corresponding to the given IDs.
    - Source - If not None, specifies the sources whose elements should be selected by ID.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
        should modify the existing selection.
    """
    if not Source:
        Source = paraview.simple.GetActiveSource()

    repr = paraview.simple.GetRepresentation(Source)

    import paraview.vtk as vtk
    reprCollection = vtk.vtkCollection()
    reprCollection.AddItem(repr.SMProxy)

    selection = _createSelection(proxyname, IDs=IDs, FieldType=FieldType, ContainingCells=ContainingCells)
    if selection:
        selectionCollection = vtk.vtkCollection()
        selectionCollection.AddItem(selection.SMProxy)

        _collectSelectionPorts(reprCollection, selectionCollection, Modifier=Modifier)

    Source.UpdateVTKObjects()

    paraview.simple.RenderAllViews()


def SelectGlobalIDs(IDs=[], FieldType='POINT', ContainingCells=False, Source=None, Modifier=None):
    """Select attributes by global IDs.

    - IDs - list of IDs of attribute types to select. Defined as a list of global IDs.
    - FieldType - type of attribute to select, e.g., 'POINT', 'CELL'
    - ContainingCells - if True and FieldType is 'POINT', select the cells containing the
        points corresponding to the given IDs.
    - Source - If not None, specifies the sources whose elements should be selected by ID.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
        should modify the existing selection.
    """
    _selectIDsHelper('GlobalIDSelectionSource', **locals())


def SelectPedigreeIDs(IDs=[], FieldType='POINT', ContainingCells=False, Source=None, Modifier=None):
    """Select attributes by Pedigree IDs.

    - IDs - list of IDs of attribute types to select. Defined as (domain, ID) pairs
      interleaved in a single list.
    - FieldType - type of attribute to select, e.g., 'POINT', 'CELL'
    - ContainingCells - if True and FieldType is 'POINT', select the cells containing the
        points corresponding to the given IDs.
    - Source - If not None, specifies the sources whose elements should be selected by ID.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
        should modify the existing selection.
    """
    _selectIDsHelper('PedigreeIDSelectionSource', **locals())


def SelectIDs(IDs=[], FieldType='POINT', ContainingCells=False, Source=None, Modifier=None):
    """Select attributes by attribute IDs.

    - IDs - list of IDs of attribute types to select. Defined as (process number, attribute ID) pairs
      interleaved in a single list. For multiblock datasets, this will select attributes on all
      blocks of the provided (processor number, attribute ID) pairs
    - FieldType - type of attribute to select, e.g., 'POINT', 'CELL'
    - ContainingCells - if True and FieldType is 'POINT', select the cells containing the
        points corresponding to the given IDs.
    - Source - If not None, specifies the sources whose elements should be selected by ID.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
        should modify the existing selection.
    """
    _selectIDsHelper('IDSelectionSource', **locals())


def SelectCompositeDataIDs(IDs=[], FieldType='POINT', ContainingCells=False, Source=None, Modifier=None):
    """Select attributes by composite attribute IDs.

    - IDs - list of IDs of attribute types to select. Defined as 3-tuples of
      (flat block index, process number, attribute ID) interleaved in a single list.
    - FieldType - type of attribute to select, e.g., 'POINT', 'CELL'
    - ContainingCells - if True and FieldType is 'POINT', select the cells containing the
        points corresponding to the given IDs.
    - Source - If not None, specifies the sources whose elements should be selected by ID.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
        should modify the existing selection.
    """
    _selectIDsHelper('CompositeDataIDSelectionSource', **locals())


def SelectHierarchicalDataIDs(IDs=[], FieldType='POINT', ContainingCells=False, Source=None, Modifier=None):
    """Select attributes by hierarchical data IDs.

    - IDs - list of IDs of attribute types to select. Defined as 3-tuples of
      (level, index, attribute ID) interleaved in a single list.
    - FieldType - type of attribute to select, e.g., 'POINT', 'CELL'
    - ContainingCells - if True and FieldType is 'POINT', select the cells containing the
        points corresponding to the given IDs.
    - Source - If not None, specifies the sources whose elements should be selected by ID.
    - Modifier - 'ADD', 'SUBTRACT', 'TOGGLE', or None to define whether and how the selection
        should modify the existing selection.
    """
    _selectIDsHelper('HierarchicalDataIDSelectionSource', **locals())


def SelectThresholds(Thresholds=[], ArrayName='', FieldType='POINT', Source=None, Modifier=None):
    """Select attributes in a source by thresholding on values in an associated array.

    - Thresholds - list of lower and upper threshold bounds. Attributes with associated
      values in the selected array between any set of bounds will be selected.
    - ArrayName - name of the array to threshold.
    - FieldType - atttribute to select, e.g., 'POINT' or 'CELL'
    - Source - if not set, then the selection will be on the active source
    """
    if not Source:
        Source = paraview.simple.GetActiveSource()

    repr = paraview.simple.GetRepresentation(Source)

    import paraview.vtk as vtk
    reprCollection = vtk.vtkCollection()
    reprCollection.AddItem(repr.SMProxy)

    selection = _createSelection('ThresholdSelectionSource', Thresholds=Thresholds, ArrayName=ArrayName,
                                 FieldType=FieldType)
    if selection:
        selectionCollection = vtk.vtkCollection()
        selectionCollection.AddItem(selection.SMProxy)

        _collectSelectionPorts(reprCollection, selectionCollection, Modifier=Modifier)

    Source.UpdateVTKObjects()


def SelectLocation(Locations=[], Source=None, Modifier=None):
    """Select points by location.

    - Locations - list of x, y, z points to select.
    - Source - if not set, then the selection will be on the active source
    """
    if not Source:
        Source = paraview.simple.GetActiveSource()

    repr = paraview.simple.GetRepresentation(Source)

    import paraview.vtk as vtk
    reprCollection = vtk.vtkCollection()
    reprCollection.AddItem(repr.SMProxy)

    selection = _createSelection('LocationSelectionSource', Locations=Locations, FieldType='POINT')
    if selection:
        selectionCollection = vtk.vtkCollection()
        selectionCollection.AddItem(selection.SMProxy)

        _collectSelectionPorts(reprCollection, selectionCollection, Modifier=Modifier)

    Source.UpdateVTKObjects()


def QuerySelect(QueryString='', FieldType='POINT', Source=None, InsideOut=False):
    """
    Selection by query expression.

    :param QueryString: string with NumPy-like boolean expression defining which attributes are selected
    :type QueryString: str
    :param FieldType: attribute to select, e.g., 'POINT' or 'CELL'
    :type FieldType: str
    :param Source: if not set, then the selection will be on the active source
    :type Source: paraview.servermanager.source.Proxy
    :param InsideOut: Invert the selection so that attributes that do not satisfy the
        expression are selected instead of elements that do.
    :type InsideOut: bool
    :return: None

    """
    if not Source:
        Source = paraview.simple.GetActiveSource()

    repr = paraview.simple.GetRepresentation(Source)

    import paraview.vtk as vtk
    reprCollection = vtk.vtkCollection()
    reprCollection.AddItem(repr.SMProxy)

    # convert FieldType to ElementType. Eventually, all public API should change
    # to accepting ElementType but we'll do that after all selection sources use
    # ElementType consistently.
    ftype = vtk.vtkSelectionNode.GetFieldTypeFromString(FieldType)
    if ftype == vtk.vtkSelectionNode.NUM_FIELD_TYPES:
        raise ValueError("Invalid FieldType '%s'" % FieldType)
    ElementType = vtk.vtkSelectionNode.ConvertSelectionFieldToAttributeType(ftype)

    selection = _createSelection('SelectionQuerySource', ElementType=ElementType,
                                 QueryString=QueryString, InsideOut=InsideOut)
    if selection:
        selectionCollection = vtk.vtkCollection()
        selectionCollection.AddItem(selection.SMProxy)

        _collectSelectionPorts(reprCollection, selectionCollection)

    Source.UpdateVTKObjects()


def ClearSelection(Source=None):
    """ Clears the selection on the source passed in to the source parameter
        or the active source if no source is provided.
    """
    if Source == None:
        Source = paraview.simple.GetActiveSource()

    Source.SMProxy.SetSelectionInput(0, None, 0)


def GrowSelection(Source=None, Layers=1):
    """Grows a selection by a number of layers.

    :param Source: If provided, the source whose selection should be modified. Defaults to the active source.
    :type Source: Source proxy
    :param Layers: The number of layers of points (or cells depending on the selection type) by which to grow
        the selection. Can be negative to shrink the selection. Defaults to 1.
    :type Layers: int
    :rtype: None"""
    if Source == None:
        Source = paraview.simple.GetActiveSource()

    if Source == None or not Source.SMProxy.IsA("vtkSMSourceProxy"):
        return

    selection_source = sm.vtkSMPropertyHelper(Source.SMProxy.GetSelectionInput(0), "Input").GetAsProxy()
    settings = sm.vtkPVRenderViewSettings.GetInstance();
    remove_seed = settings.GetGrowSelectionRemoveSeed()
    remove_intermediate_layers = settings.GetGrowSelectionRemoveIntermediateLayers()
    sm.vtkSMSelectionHelper.ExpandSelection(selection_source, Layers, remove_seed, remove_intermediate_layers)


def ShrinkSelection(Source=None, Layers=1):
    """Shrinks a selection by a number of layers.

    :param Source: If provided, the source whose selection should be modified. Defaults to the active source.
    :type Source: Source proxy
    :param Layers: The number of layers of points (or cells depending on the selection type) by which to shrink
        the selection. Can be negative to grow the selection. Defaults to 1.
    :type Layers: int
    :rtype: None"""
    GrowSelection(Source, -Layers)



import paraview

VERSION = paraview.compatibility.GetVersion()


def GetPaletteName(paletteName):
    name = paletteName

    if VERSION <= (5, 10):
        if paletteName == "DefaultBackground" or paletteName == "GrayBackground":
            name = "BlueGrayBackground"
        elif paletteName == "PrintBackground":
            name = "WhiteBackground"
    if VERSION <= (5, 11):
        if paletteName == "WarmGrayBackground":
            name = "DarkGrayBackground"
    if VERSION <= (5, 12):
        if paletteName == "DefaultBackground":
            name = "BlueGrayBackground"

    return name



#
# layer class
#
class layer:
    """Composible Image Set layer class

    A layer contains:
    - a set of channels.
        - one channel may be depth
        - one channel may be shadow
    - dimensions: integer width and height.
    - offset: integer x, y dimension from the image origin.
    """

    def __init__(self, name):
        self.depth = None
        self.shadow = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value):
        self._channel = value

    @property
    def dims(self):
        return self._dims

    @dims.setter
    def dims(self, value):
        self._dims = value

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = value

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value):
        self._depth = value

    @property
    def shadow(self):
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        self._shadow = value



from paraview import servermanager

from paraview.selection import (
    SelectionProxy,
    CreateSelection,
    SelectSurfacePoints,
    SelectSurfaceCells,
    SelectSurfaceBlocks,
    SelectPointsThrough,
    SelectCellsThrough,
    SelectGlobalIDs,
    SelectPedigreeIDs,
    SelectIDs,
    SelectCompositeDataIDs,
    SelectHierarchicalDataIDs,
    SelectThresholds,
    SelectLocation,
    QuerySelect,
)
from paraview.simple.session import GetActiveSource

__all__ = [
    "SelectionProxy",
    "CreateSelection",
    "SelectSurfacePoints",
    "SelectSurfaceCells",
    "SelectSurfaceBlocks",
    "SelectPointsThrough",
    "SelectCellsThrough",
    "SelectGlobalIDs",
    "SelectPedigreeIDs",
    "SelectIDs",
    "SelectCompositeDataIDs",
    "SelectHierarchicalDataIDs",
    "SelectThresholds",
    "SelectLocation",
    "QuerySelect",
    "SelectCells",
    "SelectPoints",
    "ClearSelection",
]


# ==============================================================================
# Selection Management
# ==============================================================================
def _select(seltype, query=None, proxy=None):
    """Function to help create query selection sources.

    :param seltype: Selection type
    :type seltype: str
    :param query: Query expression that defines the selection. Optional, defaults
        to expression that selects all cells.
    :type query: str
    :param proxy: The source proxy to select from. Optional, defaults to the
        active source.
    :type proxy: Source proxy.
    :return: The created selection source.
    :rtype: :class:`servermanager.sources.SelectionQuerySource`"""
    if not proxy:
        proxy = GetActiveSource()
    if not proxy:
        raise RuntimeError("No active source was found.")

    if not query:
        # This ends up being true for all cells.
        query = "id >= 0"

    from paraview.vtk import vtkDataObject

    # Note, selSource is not registered with the proxy manager.
    selSource = servermanager.sources.SelectionQuerySource()
    if seltype.lower() == "point":
        elementType = vtkDataObject.POINT
    elif seltype.lower() == "cell":
        elementType = vtkDataObject.CELL
    else:
        elementType = seltype
    selSource.ElementType = elementType
    selSource.QueryString = str(query)
    proxy.SMProxy.SetSelectionInput(proxy.Port, selSource.SMProxy, 0)
    return selSource


# -----------------------------------------------------------------------------


def SelectCells(query=None, proxy=None):
    """Select cells satisfying the query.

    :param query: The selection query. If `None`, then all cells are
        selected.
    :type query: str
    :param proxy: The source proxy to select from. Optional, defaults to the
        active source.
    :type proxy: Source proxy
    :return: Selection source
    :rtype: :class:`servermanager.sources.SelectionQuerySource`"""
    return _select("CELL", query, proxy)


# -----------------------------------------------------------------------------


def SelectPoints(query=None, proxy=None):
    """Select points satisfying the query.

    :param query: The selection query. If `None`, then all points are
        selected.
    :type query: str
    :param proxy: The source proxy to select from. Optional, defaults to the
        active source.
    :type proxy: Source proxy
    :return: Selection source
    :rtype: :class:`servermanager.sources.SelectionQuerySource`"""
    return _select("POINT", query, proxy)


# -----------------------------------------------------------------------------


def ClearSelection(proxy=None):
    """Clears the selection on the active source.

    :param proxy: Source proxy whose selection should be cleared.
    :type proxy: Source proxy"""
    if not proxy:
        proxy = GetActiveSource()
    if not proxy:
        raise RuntimeError("No active source was found.")
    proxy.SMProxy.SetSelectionInput(proxy.Port, None, 0)



import paraview
from . import logger

from vtkmodules.vtkCommonCore import *
from vtkmodules.vtkCommonDataModel import *
from vtkmodules.vtkCommonExecutionModel import *
from vtkmodules.vtkFiltersProgrammable import *
from vtkmodules.vtkParallelCore import *
from vtkmodules.vtkCommonMisc import *
from vtkmodules.vtkCommonMath import *
from vtkmodules.vtkCommonSystem import *
from vtkmodules.vtkCommonTransforms import *

# now import modules that need not be present
try:
    from vtkmodules.vtkCommonComputationalGeometry import *
except ImportError:
    logger.debug("Error: Could not import vtkCommonComputationalGeometry")
try:
    from vtkmodules.vtkRenderingCore import vtkCamera
except ImportError:
    logger.debug("Error: Could not import vtkRenderingCore")
try:
    from vtkmodules.vtkFiltersCore import *
except ImportError:
    logger.debug("Error: Could not import vtkFiltersCore")



from vtkmodules.web import camera, javascriptMapping, arrayTypesMapping
from paraview.web import data_writer, data_converter

from vtkmodules.web.query_data_model import DataHandler
from paraview.web.camera import update_camera
from vtkmodules.web import iteritems
from paraview import simple
from paraview import servermanager

from vtkmodules.vtkCommonCore import (
    vtkFloatArray,
    vtkIdList,
    vtkUnsignedCharArray,
    vtkTypeUInt32Array,
)
from vtkmodules.vtkCommonDataModel import vtkDataSetAttributes

import json, os, math, gzip, shutil, hashlib

# Global helper variables
encode_codes = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

# -----------------------------------------------------------------------------
# Basic Dataset Builder
# -----------------------------------------------------------------------------


class DataSetBuilder(object):
    def __init__(self, location, camera_data, metadata={}, sections={}):
        self.dataHandler = DataHandler(location)
        self.cameraDescription = camera_data
        self.camera = None

        for key, value in iteritems(metadata):
            self.dataHandler.addMetaData(key, value)

        for key, value in iteritems(sections):
            self.dataHandler.addSection(key, value)

        # Update the can_write flag for MPI
        self.dataHandler.can_write = (
            servermanager.vtkProcessModule.GetProcessModule().GetPartitionId() == 0
        )

    def getDataHandler(self):
        return self.dataHandler

    def getCamera(self):
        return self.camera

    def updateCamera(self, camera):
        update_camera(self.view, camera)

    def start(self, view=None):
        if view:
            # Keep track of the view
            self.view = view

            # Handle camera if any
            if self.cameraDescription:
                if self.cameraDescription["type"] == "spherical":
                    self.camera = camera.SphericalCamera(
                        self.dataHandler,
                        view.CenterOfRotation,
                        view.CameraPosition,
                        view.CameraViewUp,
                        self.cameraDescription["phi"],
                        self.cameraDescription["theta"],
                    )
                elif self.cameraDescription["type"] == "cylindrical":
                    self.camera = camera.CylindricalCamera(
                        self.dataHandler,
                        view.CenterOfRotation,
                        view.CameraPosition,
                        view.CameraViewUp,
                        self.cameraDescription["phi"],
                        self.cameraDescription["translation"],
                    )
                elif self.cameraDescription["type"] == "cube":
                    self.camera = camera.CubeCamera(
                        self.dataHandler,
                        self.cameraDescription["viewForward"],
                        self.cameraDescription["viewUp"],
                        self.cameraDescription["positions"],
                    )
                elif self.cameraDescription["type"] == "cube-stereo":
                    self.camera = camera.StereoCubeCamera(
                        self.dataHandler,
                        self.cameraDescription["viewForward"],
                        self.cameraDescription["viewUp"],
                        self.cameraDescription["positions"],
                        self.cameraDescription["eyeSeparation"],
                    )

            # Update background color
            bgColor = view.Background
            bgColorString = "rgb(%d, %d, %d)" % tuple(
                int(bgColor[i] * 255) for i in range(3)
            )

            if view.UseGradientBackground:
                bgColor2 = view.Background2
                bgColor2String = "rgb(%d, %d, %d)" % tuple(
                    int(bgColor2[i] * 255) for i in range(3)
                )
                self.dataHandler.addMetaData(
                    "backgroundColor",
                    "linear-gradient(%s,%s)" % (bgColor2String, bgColorString),
                )
            else:
                self.dataHandler.addMetaData("backgroundColor", bgColorString)

        # Update file patterns
        self.dataHandler.updateBasePattern()

    def stop(self):
        self.dataHandler.writeDataDescriptor()


# -----------------------------------------------------------------------------
# Image Dataset Builder
# -----------------------------------------------------------------------------


class ImageDataSetBuilder(DataSetBuilder):
    def __init__(self, location, imageMimeType, cameraInfo, metadata={}):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata)
        imageExtenstion = "." + imageMimeType.split("/")[1]
        self.dataHandler.registerData(
            name="image", type="blob", mimeType=imageMimeType, fileName=imageExtenstion
        )

    def writeImages(self):
        for cam in self.camera:
            update_camera(self.view, cam)
            simple.SaveScreenshot(self.dataHandler.getDataAbsoluteFilePath("image"))


# -----------------------------------------------------------------------------
# Data Prober Dataset Builder
# -----------------------------------------------------------------------------
class DataProberDataSetBuilder(DataSetBuilder):
    def __init__(
        self,
        input,
        location,
        sampling_dimesions,
        fields_to_keep,
        custom_probing_bounds=None,
        metadata={},
    ):
        DataSetBuilder.__init__(self, location, None, metadata)
        self.fieldsToWrite = fields_to_keep
        self.resamplerFilter = simple.ResampleToImage(Input=input)
        self.resamplerFilter.SamplingDimensions = sampling_dimesions
        if custom_probing_bounds:
            self.resamplerFilter.UseInputBounds = 0
            self.resamplerFilter.SamplingBounds = custom_probing_bounds
        else:
            self.resamplerFilter.UseInputBounds = 1

        # Register all fields
        self.dataHandler.addTypes("data-prober", "binary")
        self.DataProber = {
            "types": {},
            "dimensions": sampling_dimesions,
            "ranges": {},
            "spacing": [1, 1, 1],
        }
        for field in self.fieldsToWrite:
            self.dataHandler.registerData(
                name=field, type="array", rootFile=True, fileName="%s.array" % field
            )

    def writeData(self, time=0):
        if not self.dataHandler.can_write:
            return

        self.resamplerFilter.UpdatePipeline(time)
        imageData = self.resamplerFilter.GetClientSideObject().GetOutput()
        self.DataProber["spacing"] = imageData.GetSpacing()
        arrays = imageData.GetPointData()
        maskArray = arrays.GetArray(vtkDataSetAttributes.GhostArrayName())
        for field in self.fieldsToWrite:
            array = arrays.GetArray(field)
            if array:
                if array.GetNumberOfComponents() == 1:
                    # Push NaN when no value are present instead of 0
                    for idx in range(maskArray.GetNumberOfTuples()):
                        if maskArray.GetValue(idx) == 2:  # Hidden point
                            array.SetValue(idx, float("NaN"))

                    with open(
                        self.dataHandler.getDataAbsoluteFilePath(field), "wb"
                    ) as f:
                        f.write(memoryview(array))

                    self.expandRange(array)
                else:
                    magarray = array.NewInstance()
                    magarray.SetNumberOfTuples(array.GetNumberOfTuples())
                    magarray.SetName(field)

                    for idx in range(magarray.GetNumberOfTuples()):
                        if maskArray.GetValue(idx) == 2:  # Hidden point
                            # Push NaN when no value are present
                            magarray.SetValue(idx, float("NaN"))
                        else:
                            entry = array.GetTuple(idx)
                            mag = self.magnitude(entry)
                            magarray.SetValue(idx, mag)

                    with open(
                        self.dataHandler.getDataAbsoluteFilePath(field), "wb"
                    ) as f:
                        f.write(memoryview(magarray))

                    self.expandRange(magarray)
            else:
                print("No array for", field)
                print(self.resamplerFilter.GetOutput())

    def magnitude(self, tuple):
        value = 0
        for item in tuple:
            value += item * item
        value = value ** 0.5

        return value

    def expandRange(self, array):
        field = array.GetName()
        self.DataProber["types"][field] = javascriptMapping[
            arrayTypesMapping[array.GetDataType()]
        ]

        if field in self.DataProber["ranges"]:
            dataRange = array.GetRange()
            if dataRange[0] < self.DataProber["ranges"][field][0]:
                self.DataProber["ranges"][field][0] = dataRange[0]
            if dataRange[1] > self.DataProber["ranges"][field][1]:
                self.DataProber["ranges"][field][1] = dataRange[1]
        else:
            self.DataProber["ranges"][field] = [
                array.GetRange()[0],
                array.GetRange()[1],
            ]

    def stop(self, compress=True):
        # Rescale spacing to have the smaller value to be 1.0
        smallerValue = min(self.DataProber["spacing"])
        if smallerValue < 1.0:
            self.DataProber["spacing"] = tuple(
                i / smallerValue for i in self.DataProber["spacing"]
            )

        # Push metadata
        self.dataHandler.addSection("DataProber", self.DataProber)

        # Write metadata
        DataSetBuilder.stop(self)

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print("Compress", root)
                for name in files:
                    if ".array" in name and ".gz" not in name:
                        with open(os.path.join(root, name), "rb") as f_in:
                            with gzip.open(
                                os.path.join(root, name + ".gz"), "wb"
                            ) as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))


# -----------------------------------------------------------------------------
# Float Image with Layer Dataset Builder
# -----------------------------------------------------------------------------


class LayerDataSetBuilder(DataSetBuilder):
    def __init__(self, input, location, cameraInfo, imageSize=[500, 500], metadata={}):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata)
        self.dataRenderer = data_writer.ScalarRenderer(
            isWriter=self.dataHandler.can_write
        )
        self.view = self.dataRenderer.getView()
        self.view.ViewSize = imageSize
        self.floatImage = {"dimensions": imageSize, "layers": [], "ranges": {}}
        self.layerMap = {}
        self.input = input
        self.activeLayer = None
        self.activeField = None
        self.layerChanged = False
        self.lastTime = -1

        # Update data type
        self.dataHandler.addTypes("float-image")

    def getView(self):
        return self.view

    def setActiveLayer(self, layer, field, hasMesh=False, activeSource=None):
        if activeSource:
            self.activeSource = activeSource
        else:
            self.activeSource = self.input
        needDataRegistration = False
        if layer not in self.layerMap:
            layerObj = {
                "name": layer,
                "array": field,
                "arrays": [field],
                "active": True,
                "type": "Float32Array",
                "hasMesh": hasMesh,
            }
            self.layerMap[layer] = layerObj
            self.floatImage["layers"].append(layerObj)
            needDataRegistration = True

            # Register layer lighting
            self.dataHandler.registerData(
                name="%s__light" % layer,
                type="array",
                rootFile=True,
                fileName="%s__light.array" % layer,
                categories=["%s__light" % layer],
            )

            # Register layer mesh
            if hasMesh:
                self.dataHandler.registerData(
                    name="%s__mesh" % layer,
                    type="array",
                    rootFile=True,
                    fileName="%s__mesh.array" % layer,
                    categories=["%s__mesh" % layer],
                )

        elif field not in self.layerMap[layer]["arrays"]:
            self.layerMap[layer]["arrays"].append(field)
            needDataRegistration = True

        # Keep track of the active data
        if self.activeLayer != layer:
            self.layerChanged = True
        self.activeLayer = layer
        self.activeField = field

        if needDataRegistration:
            self.dataHandler.registerData(
                name="%s_%s" % (layer, field),
                type="array",
                rootFile=True,
                fileName="%s_%s.array" % (layer, field),
                categories=["%s_%s" % (layer, field)],
            )

    def writeLayerData(self, time=0):
        dataRange = [0, 1]
        self.activeSource.UpdatePipeline(time)

        if self.activeField and self.activeLayer:

            if self.layerChanged or self.lastTime != time:
                self.layerChanged = False
                self.lastTime = time

                # Capture lighting information
                for camPos in self.getCamera():
                    self.view.CameraFocalPoint = camPos["focalPoint"]
                    self.view.CameraPosition = camPos["position"]
                    self.view.CameraViewUp = camPos["viewUp"]
                    self.dataRenderer.writeLightArray(
                        self.dataHandler.getDataAbsoluteFilePath(
                            "%s__light" % self.activeLayer
                        ),
                        self.activeSource,
                    )

                # Capture mesh information
                if self.layerMap[self.activeLayer]["hasMesh"]:
                    for camPos in self.getCamera():
                        self.view.CameraFocalPoint = camPos["focalPoint"]
                        self.view.CameraPosition = camPos["position"]
                        self.view.CameraViewUp = camPos["viewUp"]
                        self.dataRenderer.writeMeshArray(
                            self.dataHandler.getDataAbsoluteFilePath(
                                "%s__mesh" % self.activeLayer
                            ),
                            self.activeSource,
                        )

            for camPos in self.getCamera():
                self.view.CameraFocalPoint = camPos["focalPoint"]
                self.view.CameraPosition = camPos["position"]
                self.view.CameraViewUp = camPos["viewUp"]
                dataName = "%s_%s" % (self.activeLayer, self.activeField)
                dataRange = self.dataRenderer.writeArray(
                    self.dataHandler.getDataAbsoluteFilePath(dataName),
                    self.activeSource,
                    self.activeField,
                )

            if self.activeField not in self.floatImage["ranges"]:
                self.floatImage["ranges"][self.activeField] = [
                    dataRange[0],
                    dataRange[1],
                ]
            else:
                # Expand the ranges
                if dataRange[0] < self.floatImage["ranges"][self.activeField][0]:
                    self.floatImage["ranges"][self.activeField][0] = dataRange[0]
                if dataRange[1] > self.floatImage["ranges"][self.activeField][1]:
                    self.floatImage["ranges"][self.activeField][1] = dataRange[1]

    def start(self):
        DataSetBuilder.start(self, self.view)

    def stop(self, compress=True):
        if not self.dataHandler.can_write:
            return

        # Push metadata
        self.dataHandler.addSection("FloatImage", self.floatImage)

        # Write metadata
        DataSetBuilder.stop(self)

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print("Compress", root)
                for name in files:
                    if ".array" in name and ".gz" not in name:
                        with open(os.path.join(root, name), "rb") as f_in:
                            with gzip.open(
                                os.path.join(root, name + ".gz"), "wb"
                            ) as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))


# -----------------------------------------------------------------------------
# Composite Dataset Builder
# -----------------------------------------------------------------------------


class CompositeDataSetBuilder(DataSetBuilder):
    def __init__(
        self, location, sceneConfig, cameraInfo, metadata={}, sections={}, view=None
    ):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata, sections)

        if view:
            self.view = view
        else:
            self.view = simple.CreateView("RenderView")

        self.view.ViewSize = sceneConfig["size"]
        self.view.CenterAxesVisibility = 0
        self.view.OrientationAxesVisibility = 0
        self.view.UpdatePropertyInformation()
        if "camera" in sceneConfig and "CameraFocalPoint" in sceneConfig["camera"]:
            self.view.CenterOfRotation = sceneConfig["camera"]["CameraFocalPoint"]

        # Initialize camera
        for key, value in iteritems(sceneConfig["camera"]):
            self.view.GetProperty(key).SetData(value)

        # Create a representation for all scene sources
        self.config = sceneConfig
        self.representations = []
        for data in self.config["scene"]:
            rep = simple.Show(data["source"], self.view)
            self.representations.append(rep)
            if "representation" in data:
                for key in data["representation"]:
                    rep.GetProperty(key).SetData(data["representation"][key])

        # Add directory path
        self.dataHandler.registerData(
            name="directory", rootFile=True, fileName="file.txt", categories=["trash"]
        )

    def start(self):
        DataSetBuilder.start(self, self.view)

    def stop(self, compress=True, clean=True):
        DataSetBuilder.stop(self)

        if not self.dataHandler.can_write:
            return

        # Make the config serializable
        for item in self.config["scene"]:
            del item["source"]

        # Write the scene to disk
        with open(
            os.path.join(self.dataHandler.getBasePath(), "config.json"), "w"
        ) as f:
            f.write(json.dumps(self.config))

        dataConverter = data_converter.ConvertCompositeDataToSortedStack(
            self.dataHandler.getBasePath()
        )
        dataConverter.convert()

        # Remove tmp files
        os.remove(os.path.join(self.dataHandler.getBasePath(), "index.json"))
        os.remove(os.path.join(self.dataHandler.getBasePath(), "config.json"))

        # Composite pipeline meta description
        compositePipeline = {
            "default_pipeline": "",
            "layers": [],
            "fields": {},
            "layer_fields": {},
            "pipeline": [],
        }
        rootItems = {}
        fieldNameMapping = {}

        # Clean scene in config and gather ranges
        dataRanges = {}
        layerIdx = 0
        for layer in self.config["scene"]:
            # Create group node if any
            if "parent" in layer and layer["parent"] not in rootItems:
                rootItems[layer["parent"]] = {
                    "name": layer["parent"],
                    "ids": [],
                    "children": [],
                }
                compositePipeline["pipeline"].append(rootItems[layer["parent"]])

            # Create layer entry
            layerCode = encode_codes[layerIdx]
            layerItem = {"name": layer["name"], "ids": [layerCode]}
            compositePipeline["layers"].append(layerCode)
            compositePipeline["layer_fields"][layerCode] = []
            compositePipeline["default_pipeline"] += layerCode

            # Register layer entry in pipeline
            if "parent" in layer:
                rootItems[layer["parent"]]["children"].append(layerItem)
                rootItems[layer["parent"]]["ids"].append(layerCode)
            else:
                compositePipeline["pipeline"].append(layerItem)

            # Handle color / field
            colorByList = []
            for color in layer["colors"]:
                # Find color code
                if color not in fieldNameMapping:
                    colorCode = encode_codes[len(fieldNameMapping)]
                    fieldNameMapping[color] = colorCode
                    compositePipeline["fields"][colorCode] = color
                else:
                    colorCode = fieldNameMapping[color]

                # Register color code
                compositePipeline["layer_fields"][layerCode].append(colorCode)
                if len(colorByList) == 0:
                    compositePipeline["default_pipeline"] += colorCode

                values = None
                if "constant" in layer["colors"][color]:
                    value = layer["colors"][color]["constant"]
                    values = [value, value]
                    colorByList.append({"name": color, "type": "const", "value": value})
                elif "range" in layer["colors"][color]:
                    values = layer["colors"][color]["range"]
                    colorByList.append({"name": color, "type": "field"})

                if values:
                    if color not in dataRanges:
                        dataRanges[color] = values
                    else:
                        dataRanges[color][0] = min(
                            dataRanges[color][0], values[0], values[1]
                        )
                        dataRanges[color][1] = max(
                            dataRanges[color][1], values[0], values[1]
                        )

            layer["colorBy"] = colorByList
            del layer["colors"]
            layerIdx += 1

        sortedCompositeSection = {
            "dimensions": self.config["size"],
            "pipeline": self.config["scene"],
            "ranges": dataRanges,
            "layers": len(self.config["scene"]),
            "light": self.config["light"],
        }
        self.dataHandler.addSection("SortedComposite", sortedCompositeSection)
        self.dataHandler.addSection("CompositePipeline", compositePipeline)
        self.dataHandler.addTypes("sorted-composite", "multi-color-by")

        self.dataHandler.removeData("directory")
        for dataToRegister in dataConverter.listData():
            self.dataHandler.registerData(**dataToRegister)

        self.dataHandler.writeDataDescriptor()

        if clean:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print("Clean", root)
                for name in files:
                    if name in ["camera.json"]:
                        os.remove(os.path.join(root, name))

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print("Compress", root)
                for name in files:
                    if (".float32" in name or ".uint8" in name) and ".gz" not in name:
                        with open(os.path.join(root, name), "rb") as f_in:
                            with gzip.open(
                                os.path.join(root, name + ".gz"), "wb"
                            ) as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))

    def writeData(self):
        composite_size = len(self.representations)

        self.view.UpdatePropertyInformation()
        self.view.Background = [0, 0, 0]
        imageSize = self.view.ViewSize[0] * self.view.ViewSize[1]

        # Generate the heavy data
        for camPos in self.getCamera():
            self.view.CameraFocalPoint = camPos["focalPoint"]
            self.view.CameraPosition = camPos["position"]
            self.view.CameraViewUp = camPos["viewUp"]

            # Show all representations
            for compositeIdx in range(composite_size):
                rep = self.representations[compositeIdx]
                rep.Visibility = 1

            # Fix camera bounds
            self.view.LockBounds = 0
            simple.Render(self.view)
            self.view.LockBounds = 1

            # Update destination directory
            dest_path = os.path.dirname(
                self.dataHandler.getDataAbsoluteFilePath("directory")
            )

            # Write camera information
            if self.dataHandler.can_write:
                with open(os.path.join(dest_path, "camera.json"), "w") as f:
                    f.write(json.dumps(camPos))

            # Hide all representations
            for compositeIdx in range(composite_size):
                rep = self.representations[compositeIdx]
                rep.Visibility = 0

            # Show only active Representation
            # Extract images for each fields
            for compositeIdx in range(composite_size):
                rep = self.representations[compositeIdx]
                if compositeIdx > 0:
                    self.representations[compositeIdx - 1].Visibility = 0
                rep.Visibility = 1

                # capture Z
                simple.Render()
                zBuffer = self.view.CaptureDepthBuffer()
                with open(
                    os.path.join(dest_path, "depth_%d.float32" % compositeIdx), "wb"
                ) as f:
                    f.write(memoryview(zBuffer))

                # Prevent color interference
                rep.DiffuseColor = [1, 1, 1]

                # Handle light
                for lightType in self.config["light"]:
                    if lightType == "intensity":
                        rep.AmbientColor = [1, 1, 1]
                        rep.SpecularColor = [1, 1, 1]

                        self.view.StartCaptureLuminance()
                        image = self.view.CaptureWindow(1)
                        imagescalars = image.GetPointData().GetScalars()
                        self.view.StopCaptureLuminance()

                        # Extract specular information
                        specularOffset = 1  # [diffuse, specular, ?]
                        imageSize = imagescalars.GetNumberOfTuples()
                        specularComponent = vtkUnsignedCharArray()
                        specularComponent.SetNumberOfComponents(1)
                        specularComponent.SetNumberOfTuples(imageSize)

                        for idx in range(imageSize):
                            specularComponent.SetValue(
                                idx, imagescalars.GetValue(idx * 3 + specularOffset)
                            )

                        with open(
                            os.path.join(
                                dest_path, "intensity_%d.uint8" % compositeIdx
                            ),
                            "wb",
                        ) as f:
                            f.write(memoryview(specularComponent))

                        # Free memory
                        image.UnRegister(None)

                    if lightType == "normal":
                        if rep.Representation in [
                            "Point Gaussian",
                            "Points",
                            "Outline",
                            "Wireframe",
                        ]:
                            uniqNormal = [
                                (camPos["position"][i] - camPos["focalPoint"][i])
                                for i in range(3)
                            ]
                            tmpNormalArray = vtkFloatArray()
                            tmpNormalArray.SetNumberOfComponents(1)
                            tmpNormalArray.SetNumberOfTuples(imageSize)

                            for comp in range(3):
                                tmpNormalArray.FillComponent(0, uniqNormal[comp])
                                with open(
                                    os.path.join(
                                        dest_path,
                                        "normal_%d_%d.float32" % (compositeIdx, comp),
                                    ),
                                    "wb",
                                ) as f:
                                    f.write(memoryview(tmpNormalArray))
                        else:
                            for comp in range(3):
                                # Configure view to handle POINT_DATA / CELL_DATA
                                self.view.DrawCells = 0
                                self.view.ArrayNameToDraw = "Normals"
                                self.view.ArrayComponentToDraw = comp
                                self.view.ScalarRange = [-1.0, 1.0]
                                self.view.StartCaptureValues()
                                image = self.view.CaptureWindow(1)
                                imagescalars = image.GetPointData().GetScalars()
                                self.view.StopCaptureValues()

                                # Convert RGB => Float => Write
                                floatArray = data_converter.convertRGBArrayToFloatArray(
                                    imagescalars, [-1.0, 1.0]
                                )
                                with open(
                                    os.path.join(
                                        dest_path,
                                        "normal_%d_%d.float32" % (compositeIdx, comp),
                                    ),
                                    "wb",
                                ) as f:
                                    f.write(memoryview(floatArray))

                                # Free memory
                                image.UnRegister(None)

                # Handle color by
                for fieldName, fieldConfig in iteritems(
                    self.config["scene"][compositeIdx]["colors"]
                ):
                    if "constant" in fieldConfig:
                        # Skip nothing to render
                        continue

                    # Configure view to handle POINT_DATA / CELL_DATA
                    if fieldConfig["location"] == "POINT_DATA":
                        self.view.DrawCells = 0
                    else:
                        self.view.DrawCells = 1

                    self.view.ArrayNameToDraw = fieldName
                    self.view.ArrayComponentToDraw = 0
                    self.view.ScalarRange = fieldConfig["range"]
                    self.view.StartCaptureValues()
                    image = self.view.CaptureWindow(1)
                    imagescalars = image.GetPointData().GetScalars()
                    self.view.StopCaptureValues()

                    floatArray = data_converter.convertRGBArrayToFloatArray(
                        imagescalars, fieldConfig["range"]
                    )
                    with open(
                        os.path.join(
                            dest_path, "%d_%s.float32" % (compositeIdx, fieldName)
                        ),
                        "wb",
                    ) as f:
                        f.write(memoryview(floatArray))
                        self.dataHandler.registerData(
                            name="%d_%s" % (compositeIdx, fieldName),
                            fileName="/%d_%s.float32" % (compositeIdx, fieldName),
                            type="array",
                            categories=["%d_%s" % (compositeIdx, fieldName)],
                        )

                    # Free memory
                    image.UnRegister(None)


# -----------------------------------------------------------------------------
# VTKGeometryDataSetBuilder Dataset Builder
# -----------------------------------------------------------------------------


def writeCellArray(dataHandler, currentData, cellName, inputCellArray):
    nbValues = inputCellArray.GetNumberOfTuples()
    if nbValues == 0:
        return

    outputCells = vtkTypeUInt32Array()
    outputCells.SetNumberOfTuples(nbValues)

    for valueIdx in range(nbValues):
        outputCells.SetValue(valueIdx, inputCellArray.GetValue(valueIdx))

    iBuffer = memoryview(outputCells)
    iMd5 = hashlib.md5(iBuffer).hexdigest()
    iPath = os.path.join(dataHandler.getBasePath(), "data", "%s.Uint32Array" % iMd5)
    currentData["polys"] = "data/%s.Uint32Array" % iMd5
    with open(iPath, "wb") as f:
        f.write(iBuffer)


class VTKGeometryDataSetBuilder(DataSetBuilder):
    def __init__(self, location, sceneConfig, metadata={}, sections={}):
        DataSetBuilder.__init__(self, location, None, metadata, sections)

        # Update data type
        self.dataHandler.addTypes("vtk-geometry")

        # Create a representation for all scene sources
        self.config = sceneConfig

        # Processing pipeline
        self.surfaceExtract = None

        # Add directory path
        self.dataHandler.registerData(
            priority=0, name="scene", rootFile=True, fileName="scene.json", type="json"
        )

        # Create directory containers
        dataPath = os.path.join(location, "data")
        for p in [dataPath]:
            if not os.path.exists(p):
                os.makedirs(p)

        # Create metadata structure
        colorToCodeMap = {}
        parentNodes = {}
        pipelineMeta = {"layers": [], "pipeline": [], "layer_fields": {}, "fields": {}}
        geometryMeta = {
            "ranges": {},
            "layer_map": {},
        }
        self.ranges = geometryMeta["ranges"]
        for item in sceneConfig["scene"]:
            # Handle layer
            layerCode = encode_codes[len(pipelineMeta["layers"])]
            pipelineMeta["layers"].append(layerCode)
            geometryMeta["layer_map"][layerCode] = item["name"]

            # Handle colors
            pipelineMeta["layer_fields"][layerCode] = []
            for fieldName in item["colors"]:
                colorCode = None
                if fieldName in colorToCodeMap:
                    colorCode = colorToCodeMap[fieldName]
                else:
                    colorCode = encode_codes[len(colorToCodeMap)]
                    colorToCodeMap[fieldName] = colorCode
                    geometryMeta["ranges"][fieldName] = [
                        1,
                        -1,
                    ]  # FIXME we don't know the range

                pipelineMeta["layer_fields"][layerCode].append(colorCode)
                pipelineMeta["fields"][colorCode] = fieldName

            # Handle pipeline
            if "parent" in item:
                # Need to handle hierarchy
                if item["parent"] in parentNodes:
                    # Fill children
                    rootNode = parentNodes[item["parent"]]
                    rootNode["ids"].append(layerCode)
                    rootNode["children"].append(
                        {"name": item["name"], "ids": [layerCode]}
                    )
                else:
                    # Create root + register
                    rootNode = {
                        "name": item["parent"],
                        "ids": [layerCode],
                        "children": [{"name": item["name"], "ids": [layerCode]}],
                    }
                    parentNodes[item["parent"]] = rootNode
                    pipelineMeta["pipeline"].append(rootNode)
            else:
                # Add item info as a new pipeline node
                pipelineMeta["pipeline"].append(
                    {"name": item["name"], "ids": [layerCode]}
                )

        # Register metadata to be written in index.json
        self.dataHandler.addSection("Geometry", geometryMeta)
        self.dataHandler.addSection("CompositePipeline", pipelineMeta)

    def writeData(self, time=0):
        if not self.dataHandler.can_write:
            return

        currentScene = []
        for data in self.config["scene"]:
            currentData = {"name": data["name"], "fields": {}, "cells": {}}
            currentScene.append(currentData)
            if self.surfaceExtract:
                self.merge.Input = data["source"]
            else:
                self.merge = simple.MergeBlocks(Input=data["source"], MergePoints=0)
                self.surfaceExtract = simple.ExtractSurface(Input=self.merge)

            # Extract surface
            self.surfaceExtract.UpdatePipeline(time)
            ds = self.surfaceExtract.SMProxy.GetClientSideObject().GetOutputDataObject(
                0
            )
            originalDS = (
                data["source"].SMProxy.GetClientSideObject().GetOutputDataObject(0)
            )

            originalPoints = ds.GetPoints()

            # Points
            points = vtkFloatArray()
            nbPoints = originalPoints.GetNumberOfPoints()
            points.SetNumberOfComponents(3)
            points.SetNumberOfTuples(nbPoints)
            for idx in range(nbPoints):
                coord = originalPoints.GetPoint(idx)
                points.SetTuple3(idx, coord[0], coord[1], coord[2])

            pBuffer = memoryview(points)
            pMd5 = hashlib.md5(pBuffer).hexdigest()
            pPath = os.path.join(
                self.dataHandler.getBasePath(), "data", "%s.Float32Array" % pMd5
            )
            currentData["points"] = "data/%s.Float32Array" % pMd5
            with open(pPath, "wb") as f:
                f.write(pBuffer)

            # Handle cells
            writeCellArray(
                self.dataHandler, currentData["cells"], "verts", ds.GetVerts().GetData()
            )
            writeCellArray(
                self.dataHandler, currentData["cells"], "lines", ds.GetLines().GetData()
            )
            writeCellArray(
                self.dataHandler, currentData["cells"], "polys", ds.GetPolys().GetData()
            )
            writeCellArray(
                self.dataHandler,
                currentData["cells"],
                "strips",
                ds.GetStrips().GetData(),
            )

            # Fields
            for fieldName, fieldInfo in iteritems(data["colors"]):
                array = None
                if "constant" in fieldInfo:
                    currentData["fields"][fieldName] = fieldInfo
                    continue
                elif "POINT_DATA" in fieldInfo["location"]:
                    array = ds.GetPointData().GetArray(fieldName)
                elif "CELL_DATA" in fieldInfo["location"]:
                    array = ds.GetCellData().GetArray(fieldName)
                jsType = javascriptMapping[arrayTypesMapping[array.GetDataType()]]
                arrayRange = array.GetRange(-1)
                tupleSize = array.GetNumberOfComponents()
                arraySize = array.GetNumberOfTuples()
                if tupleSize == 1:
                    outputField = array
                else:
                    # compute magnitude
                    outputField = array.NewInstance()
                    outputField.SetNumberOfTuples(arraySize)
                    tupleIdxs = range(tupleSize)
                    for i in range(arraySize):
                        magnitude = 0
                        for j in tupleIdxs:
                            magnitude += math.pow(array.GetValue(i * tupleSize + j), 2)

                        outputField.SetValue(i, math.sqrt(magnitude))

                fBuffer = memoryview(outputField)
                fMd5 = hashlib.md5(fBuffer).hexdigest()
                fPath = os.path.join(
                    self.dataHandler.getBasePath(),
                    "data",
                    "%s_%s.%s" % (fieldName, fMd5, jsType),
                )
                with open(fPath, "wb") as f:
                    f.write(fBuffer)

                currentRange = self.ranges[fieldName]
                if currentRange[1] < currentRange[0]:
                    currentRange[0] = arrayRange[0]
                    currentRange[1] = arrayRange[1]
                else:
                    currentRange[0] = (
                        arrayRange[0]
                        if arrayRange[0] < currentRange[0]
                        else currentRange[0]
                    )
                    currentRange[1] = (
                        arrayRange[1]
                        if arrayRange[1] > currentRange[1]
                        else currentRange[1]
                    )

                currentData["fields"][fieldName] = {
                    "array": "data/%s_%s.%s" % (fieldName, fMd5, jsType),
                    "location": fieldInfo["location"],
                    "range": outputField.GetRange(),
                }

        # Write scene
        with open(self.dataHandler.getDataAbsoluteFilePath("scene"), "w") as f:
            f.write(json.dumps(currentScene, indent=4))

    def stop(self, compress=True, clean=True):
        if not self.dataHandler.can_write:
            return

        DataSetBuilder.stop(self)

        if compress:
            for dirName in ["fields", "index", "points"]:
                for root, dirs, files in os.walk(
                    os.path.join(self.dataHandler.getBasePath(), dirName)
                ):
                    print("Compress", root)
                    for name in files:
                        if "Array" in name and ".gz" not in name:
                            with open(os.path.join(root, name), "rb") as f_in:
                                with gzip.open(
                                    os.path.join(root, name + ".gz"), "wb"
                                ) as f_out:
                                    shutil.copyfileobj(f_in, f_out)
                            os.remove(os.path.join(root, name))


# -----------------------------------------------------------------------------
# GeometryDataSetBuilder Dataset Builder
# -----------------------------------------------------------------------------


class GeometryDataSetBuilder(DataSetBuilder):
    def __init__(self, location, sceneConfig, metadata={}, sections={}):
        DataSetBuilder.__init__(self, location, None, metadata, sections)

        # Update data type
        self.dataHandler.addTypes("geometry")

        # Create a representation for all scene sources
        self.config = sceneConfig

        # Processing pipeline
        self.surfaceExtract = None

        # Add directory path
        self.dataHandler.registerData(
            priority=0, name="scene", rootFile=True, fileName="scene.json", type="json"
        )

        # Create directory containers
        pointsPath = os.path.join(location, "points")
        polyPath = os.path.join(location, "index")
        colorPath = os.path.join(location, "fields")
        for p in [pointsPath, polyPath, colorPath]:
            if not os.path.exists(p):
                os.makedirs(p)

        # Create metadata structure
        colorToCodeMap = {}
        parentNodes = {}
        pipelineMeta = {"layers": [], "pipeline": [], "layer_fields": {}, "fields": {}}
        geometryMeta = {"ranges": {}, "layer_map": {}, "object_size": {}}
        self.objSize = geometryMeta["object_size"]
        for item in sceneConfig["scene"]:
            # Handle layer
            layerCode = encode_codes[len(pipelineMeta["layers"])]
            pipelineMeta["layers"].append(layerCode)
            geometryMeta["layer_map"][layerCode] = item["name"]
            geometryMeta["object_size"][item["name"]] = {"points": 0, "index": 0}

            # Handle colors
            pipelineMeta["layer_fields"][layerCode] = []
            for fieldName in item["colors"]:
                colorCode = None
                if fieldName in colorToCodeMap:
                    colorCode = colorToCodeMap[fieldName]
                else:
                    colorCode = encode_codes[len(colorToCodeMap)]
                    colorToCodeMap[fieldName] = colorCode
                    geometryMeta["ranges"][fieldName] = [
                        0,
                        1,
                    ]  # FIXME we don't know the range

                pipelineMeta["layer_fields"][layerCode].append(colorCode)
                pipelineMeta["fields"][colorCode] = fieldName

            # Handle pipeline
            if "parent" in item:
                # Need to handle hierarchy
                if item["parent"] in parentNodes:
                    # Fill children
                    rootNode = parentNodes[item["parent"]]
                    rootNode["ids"].append(layerCode)
                    rootNode["children"].append(
                        {"name": item["name"], "ids": [layerCode]}
                    )
                else:
                    # Create root + register
                    rootNode = {
                        "name": item["parent"],
                        "ids": [layerCode],
                        "children": [{"name": item["name"], "ids": [layerCode]}],
                    }
                    parentNodes[item["parent"]] = rootNode
                    pipelineMeta["pipeline"].append(rootNode)
            else:
                # Add item info as a new pipeline node
                pipelineMeta["pipeline"].append(
                    {"name": item["name"], "ids": [layerCode]}
                )

        # Register metadata to be written in index.json
        self.dataHandler.addSection("Geometry", geometryMeta)
        self.dataHandler.addSection("CompositePipeline", pipelineMeta)

    def writeData(self, time=0):
        if not self.dataHandler.can_write:
            return

        currentScene = []
        for data in self.config["scene"]:
            currentData = {"name": data["name"], "fields": {}}
            currentScene.append(currentData)
            if self.surfaceExtract:
                self.merge.Input = data["source"]
            else:
                self.merge = simple.MergeBlocks(Input=data["source"], MergePoints=0)
                self.surfaceExtract = simple.ExtractSurface(Input=self.merge)

            # Extract surface
            self.surfaceExtract.UpdatePipeline(time)
            ds = self.surfaceExtract.SMProxy.GetClientSideObject().GetOutputDataObject(
                0
            )
            originalDS = (
                data["source"].SMProxy.GetClientSideObject().GetOutputDataObject(0)
            )

            originalPoints = ds.GetPoints()

            # Points
            points = vtkFloatArray()
            nbPoints = originalPoints.GetNumberOfPoints()
            points.SetNumberOfComponents(3)
            points.SetNumberOfTuples(nbPoints)
            for idx in range(nbPoints):
                coord = originalPoints.GetPoint(idx)
                points.SetTuple3(idx, coord[0], coord[1], coord[2])

            pBuffer = memoryview(points)
            pMd5 = hashlib.md5(pBuffer).hexdigest()
            pPath = os.path.join(
                self.dataHandler.getBasePath(), "points", "%s.Float32Array" % pMd5
            )
            currentData["points"] = "points/%s.Float32Array" % pMd5
            with open(pPath, "wb") as f:
                f.write(pBuffer)

            # Polys
            poly = ds.GetPolys()
            nbCells = poly.GetNumberOfCells()
            cellLocation = 0
            idList = vtkIdList()
            topo = vtkTypeUInt32Array()
            topo.Allocate(poly.GetData().GetNumberOfTuples())

            for cellIdx in range(nbCells):
                poly.GetCell(cellLocation, idList)
                cellSize = idList.GetNumberOfIds()
                cellLocation += cellSize + 1
                if cellSize == 3:
                    topo.InsertNextValue(idList.GetId(0))
                    topo.InsertNextValue(idList.GetId(1))
                    topo.InsertNextValue(idList.GetId(2))
                elif cellSize == 4:
                    topo.InsertNextValue(idList.GetId(0))
                    topo.InsertNextValue(idList.GetId(1))
                    topo.InsertNextValue(idList.GetId(3))
                    topo.InsertNextValue(idList.GetId(1))
                    topo.InsertNextValue(idList.GetId(2))
                    topo.InsertNextValue(idList.GetId(3))
                else:
                    print("Cell size of", cellSize, "not supported")

            iBuffer = memoryview(topo)
            iMd5 = hashlib.md5(iBuffer).hexdigest()
            iPath = os.path.join(
                self.dataHandler.getBasePath(), "index", "%s.Uint32Array" % iMd5
            )
            currentData["index"] = "index/%s.Uint32Array" % iMd5
            with open(iPath, "wb") as f:
                f.write(iBuffer)

            # Grow object side
            self.objSize[data["name"]]["points"] = max(
                self.objSize[data["name"]]["points"], nbPoints
            )
            self.objSize[data["name"]]["index"] = max(
                self.objSize[data["name"]]["index"], topo.GetNumberOfTuples()
            )

            # Colors / FIXME
            for fieldName, fieldInfo in iteritems(data["colors"]):
                array = ds.GetPointData().GetArray(fieldName)
                tupleSize = array.GetNumberOfComponents()
                arraySize = array.GetNumberOfTuples()
                outputField = vtkFloatArray()
                outputField.SetNumberOfTuples(arraySize)
                if tupleSize == 1:
                    for i in range(arraySize):
                        outputField.SetValue(i, array.GetValue(i))
                else:
                    # compute magnitude
                    tupleIdxs = range(tupleSize)
                    for i in range(arraySize):
                        magnitude = 0
                        for j in tupleIdxs:
                            magnitude += math.pow(array.GetValue(i * tupleSize + j), 2)

                        outputField.SetValue(i, math.sqrt(magnitude))

                fBuffer = memoryview(outputField)
                fMd5 = hashlib.md5(fBuffer).hexdigest()
                fPath = os.path.join(
                    self.dataHandler.getBasePath(),
                    "fields",
                    "%s_%s.Float32Array" % (fieldName, fMd5),
                )
                with open(fPath, "wb") as f:
                    f.write(fBuffer)

                currentData["fields"][fieldName] = "fields/%s_%s.Float32Array" % (
                    fieldName,
                    fMd5,
                )

        # Write scene
        with open(self.dataHandler.getDataAbsoluteFilePath("scene"), "w") as f:
            f.write(json.dumps(currentScene, indent=2))

    def stop(self, compress=True, clean=True):
        if not self.dataHandler.can_write:
            return

        DataSetBuilder.stop(self)

        if compress:
            for dirName in ["fields", "index", "points"]:
                for root, dirs, files in os.walk(
                    os.path.join(self.dataHandler.getBasePath(), dirName)
                ):
                    print("Compress", root)
                    for name in files:
                        if "Array" in name and ".gz" not in name:
                            with open(os.path.join(root, name), "rb") as f_in:
                                with gzip.open(
                                    os.path.join(root, name + ".gz"), "wb"
                                ) as f_out:
                                    shutil.copyfileobj(f_in, f_out)
                            os.remove(os.path.join(root, name))



import importlib.abc, importlib.util
import os.path
import zipimport


class MetaPathFinder(importlib.abc.MetaPathFinder):

    def add_file(self, filename, modulename=None):
        if not hasattr(self, "_registered_files"):
            self._registered_files = {}

        if not os.path.exists(filename):
            raise RuntimeError("'%s' must be exist!" % filename)

        if modulename is None:
            modulename = os.path.splitext(os.path.basename(filename))[0]

        unique_modulename = modulename
        uniqifier = 1
        while unique_modulename in self._registered_files:
            unique_modulename = "%s%d" % (modulename, uniqifier)
            uniqifier += 1

        self._registered_files[unique_modulename] = filename
        return unique_modulename

    def find_spec(self, fullname, path, target=None):
        if not hasattr(self, "_registered_files"):
            return None

        if fullname in self._registered_files:
            modulename = fullname
            filename = self._registered_files[fullname]

            if os.path.splitext(os.path.basename(filename))[1].lower() == ".zip":
                loader = zipimport.zipimporter(filename)
                spec = importlib.util.spec_from_loader(modulename, loader)
            else:
                spec = importlib.util.spec_from_file_location(modulename, filename)
            return spec

        return None


_path_finder_instance = MetaPathFinder()


def add_file(filename, modulename=None):
    global _path_finder_instance
    return _path_finder_instance.add_file(filename, modulename)


def _run_once(f):
    def func(*args, **kwargs):
        if not f.has_run:
            f.has_run = True
            return f(*args, **kwargs)

    f.has_run = False
    return func


@_run_once
def install_pathfinder():
    global _path_finder_instance

    import sys
    sys.meta_path.insert(0, _path_finder_instance)



import numpy as np
from scipy.interpolate import interp1d


class Renderer:
    """Cinema CIS Renderer

    The Renderer.renderer() renders a CIS image into the final composited and/or
    shadowed image. Currently, it support image composition through depth-buffer
    and applies "simulated" shadow using (pre-computed) shadow map. NaN values
    in images and depth buffers are considered as identification of "background"
    pixels and will be colored with imageview.background. Renderer() returns a
    tuple of final RGB image, represented as a 3D numpy array and depth buffer
    represented as a 2D numpy array.
    """

    def __init__(self):
        return

    # Paste buffer 'src' to buffer 'dest' at the 'offset' assuming dest is
    # large enough.
    @staticmethod
    def paste(dest, src, offset):
        ends = np.array(offset) + np.array(src.shape[0:1])
        dest[offset[0]:ends[0], offset[1]:ends[1], :] = src
        return dest

    # Color a scalar value buffer 'scalars' by the 'colormap'
    @staticmethod
    def color(scalars, colormap):
        cmap_fn, values = Renderer.make_rgb_colormap(colormap)
        # rescale scalars to be within the range of the colormap
        if np.nanmin(scalars) == np.nanmax(scalars):
            # prevent divide by zero when scalars are the same values.
            scalars = values.min()
        else:
            scalars = (scalars - np.nanmin(scalars)) / \
                      (np.nanmax(scalars) - np.nanmin(scalars))
            scalars = values.min() + scalars * (values.max() - values.min())
        return cmap_fn(scalars)

    @staticmethod
    def make_rgb_colormap(colormap):
        assert colormap['colorspace'] == 'rgb'
        points = colormap['points']
        values = np.zeros(len(points))
        rgbs = np.zeros((len(points), 3))
        for i in range(len(points)):
            values[i] = points[i]['x']
            rgbs[i, 0] = points[i]['r']
            rgbs[i, 1] = points[i]['g']
            rgbs[i, 2] = points[i]['b']
        cmap_fn = interp1d(values, rgbs, axis=0)
        return cmap_fn, values

    @staticmethod
    def blend(dest, src, mask):
        dest[mask] = src[mask]

    @staticmethod
    def depth_composite(dest_color, dest_z, src_color, src_z):
        mask = np.nan_to_num(dest_z, nan=np.inf) > \
               np.nan_to_num(src_z, nan=np.inf)
        Renderer.blend(dest_color, src_color, mask)
        Renderer.blend(dest_z, src_z, mask)

    @staticmethod
    def render(iview):

        # FXIME: this assumes RGB rather than RGBA color
        canvas = np.full((iview.dims[0], iview.dims[1], 3), iview.background,
                         float)
        depth = np.full((iview.dims[0], iview.dims[1]), np.inf, float)

        layers = iview.get_layer_data()
        for name, layer in layers.items():
            # TODO: fix this loop
            # The imageview object is intended to provide a way of iterating 
            # over active layers, but at present this has a bug in it. For
            # this release, we simply query to determine if the layer is active
            # and that is correct.
            if iview.is_active_layer(name):
                data = layer.channel.data
                background = np.full((data.shape[0], data.shape[1], 3),
                                     iview.background, float)
                foreground = Renderer.color(data, layer.channel.colormap)
                if iview.use_shadow:
                    foreground = foreground * \
                                 layer.shadow.data[:, :, np.newaxis]
                Renderer.blend(foreground, background, np.isnan(data))

                rectangle = (
                    slice(layer.offset[0], layer.offset[0] + data.shape[0]),
                    slice(layer.offset[1], layer.offset[1] + data.shape[1]))
                if iview.use_depth:
                    Renderer.depth_composite(canvas[rectangle],
                                             depth[rectangle],
                                             foreground, layer.depth.data)
                else:
                    canvas = Renderer.paste(canvas, foreground, layer.offset)

        # iview.dims is in row by column rather than x-dims by y-dims,
        # we need to transpose the two axes.
        return canvas.transpose((1, 0, 2)), depth.T



from paraview import simple
import os, json

# -----------------------------------------------------------------------------


def getAllNames():
    actorNameMapping = {}
    srcs = simple.GetSources()
    duplicates = {}
    for key, val in srcs.items():
        # Prevent name duplication
        nameToUse = key[0]
        if nameToUse in duplicates:
            count = 1
            newName = "%s (%d)" % (nameToUse, count)
            while newName in duplicates:
                count += 1
                newName = "%s (%d)" % (nameToUse, count)
            nameToUse = newName
        duplicates[nameToUse] = True
        representation = simple.GetRepresentation(val)
        if representation:
            vtkRepInstance = representation.GetClientSideObject()
            if "GetActiveRepresentation" in dir(vtkRepInstance):
                actorRep = vtkRepInstance.GetActiveRepresentation().GetActor()
                actorNameMapping[nameToUse] = actorRep
    return actorNameMapping


# -----------------------------------------------------------------------------


def getRenameMap():
    renameMap = {}
    names = getAllNames()
    view = simple.GetActiveView()
    renderer = view.GetClientSideObject().GetRenderer()
    viewProps = renderer.GetViewProps()
    volumes = renderer.GetVolumes()
    idx = 1
    for viewProp in viewProps:
        if not viewProp.IsA("vtkActor"):
            continue
        if not viewProp.GetVisibility():
            continue
        # The mapping will fail for multiblock that are composed of several blocks
        # Merge block should be used to solve the renaming issue for now
        # as the id is based on the a valid block vs representation.
        strIdx = "%s" % idx
        # Prop is valid if we can find it in the current sources
        for name, actor in names.items():
            if viewProp == actor:
                renameMap[strIdx] = name
                idx += 1
                break

    for volume in volumes:
        if not volume.IsA("vtkVolume"):
            continue
        if not volume.GetVisibility():
            continue
        strIdx = "%s" % idx
        for name, actor in names.items():
            if viewProp == actor:
                renameMap[strIdx] = name
                idx += 1
                break

    return renameMap


# -----------------------------------------------------------------------------


def applyParaViewNaming(directoryPath):
    renameMap = getRenameMap()
    scene = None
    filePath = os.path.join(directoryPath, "index.json")
    with open(filePath) as file:
        scene = json.load(file)
        for item in scene["scene"]:
            if item["name"] in renameMap:
                item["name"] = renameMap[item["name"]]

    with open(filePath, "w") as file:
        file.write(json.dumps(scene, indent=2))
